
from FITS import FITS
from SimpleQueue import *
from basics import *

from twisted.internet import abstract, threads, task
from twisted.python import log as logging
from twisted.application import service, internet
from twisted.python.components import registerAdapter

from nevow               import inevow
from nevow               import accessors
from nevow               import appserver
from nevow               import athena
from nevow               import rend
from nevow               import loaders
from nevow               import url
from nevow               import tags as T
from nevow               import flat
from nevow               import util
from nevow               import livepage
from nevow               import canvas
from nevow               import static

from twisted.internet import task
from twisted.internet import defer, protocol, reactor, address, utils

import time, sys, os
from glob import glob

from twisted.python.util import sibpath
from twisted.python.filepath import FilePath

from CIRCEHelpers import *

sibling = FilePath(__file__).sibling

mouseover = u'this.style.cursor="pointer"; this.bgColor="yellow"'
mouseout = u'this.bgColor="white";'

def callRemote(dummy,obj,func,*args,**kws):
    return getattr(obj,"callRemote")(func,*args)

class RootPage(athena.LivePage):
    def __init__ ( self, *args, **kwargs ):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        athena.LivePage.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.jsModules.mapping[u"EchoThing"] = sibling("HubSite.js").path

    def locateChild(self,ctx, segments):
        request = inevow.IRequest(ctx)
        if len(segments)>1 and os.path.exists(self.runner.basedir+"/".join(segments)):
           I = None
           suff = segments[-1].split(".")[-1]
           if suff in ["jpg","png","gif"]:
              I = static.File( self.runner.basedir+"/".join(segments),"image/%s" % (suff.replace("jpg","jpeg")))
           elif suff in ["txt","dat"]:
              I = static.File( self.runner.basedir+"/".join(segments),"text/plain")
           elif suff == "fits":
              I = static.File( self.runner.basedir+"/".join(segments),"application/fits")
           if I:
              return I,()
           else:
              return athena.LivePage.locateChild(self,ctx,segments)
        return athena.LivePage.locateChild(self,ctx,segments)

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        self.targetsTable.disconn(reason)
        self.targetWatcher.disconn(reason)
        self.pipelineStatus.disconn(reason)
        for n in range(10):
           if hasattr(self,"watch%d"%(n+1)):
              w = getattr(self,"watch%d"%(n+1))
              if w.running: w.stop()


    def render_ticker(self,ctx,data):
        obj = FrameTicker(runner=self.runner)
        obj.setFragmentParent(self)
        self.frameTicker = obj
        return ctx.tag[obj]

    def render_status(self,ctx,data):
        obj = PipelineStatus(runner=self.runner)
        obj.setFragmentParent(self)
        self.pipelineStatus = obj
        return ctx.tag[obj]

    def render_targets(self,ctx,data):
        obj = TargetsTable(runner=self.runner)
        obj.setFragmentParent(self)
        self.targetsTable = obj
        return ctx.tag[obj]

    def render_watcher(self,ctx,data):
        obj = TargetWatcher(runner=self.runner)
        obj.setFragmentParent(self)
        self.targetWatcher=obj
        return ctx.tag[obj]

    def makeDocFactory ( self ):
        return loaders.stan(T.html[
                T.head(render=T.directive('liveglue'))[
                  T.title["Carnegie InfraRed Computational Extravaganza"]],
                  T.body[
                     T.table[
                     T.tr[
                       T.td[
                         T.table[
                         T.tbody[
                           T.tr[
                            T.td(style=u"border:thin solid black;")[T.h2(style=u'width:400')[
                                "Carnegie InfraRed Computational Extravaganza"]],
                            ],
#                           T.tr[
#                            T.td(style=u"height:160;border:thin solid black;")[T.div(render=T.directive("ticker"))],
#                            ],
                           T.tr[
                            T.td(style=u"height:30;border:thin solid black;")[T.div(render=T.directive("status"))],
                            ],
                           T.tr[
                            T.td(style=u"border:thin solid black;")[T.div(render=T.directive("targets"))],
                            ]
                         ]
                         ]
                         ],
                       T.td(style=u"height:720;border:thin solid black;")[T.div(render=T.directive("watcher"))],
                     ],
                     ],
                     ]
                ])

class CIRCEMainPage(RootPage):
    def child_( self, ctx):
        return RootPage(runner=self.runner)

class FrameTicker(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.datadirs = self.runner.datadirs
#        self.ifiles = glob(self.runner.datadir+"/image?????.fits")
        self.ifiles = []
        self.chipim = {}
        self.showing = None

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        for n in range(10):
           if hasattr(self,"watch%d"%(n+1)):
              w = getattr(self,"watch%d"%(n+1))
              if w.running: w.stop()


    def fireItUp( self):
#        self.clock = task.LoopingCall(self.loop)
#        self.clock.start(1)
        reactor.callInThread(self.loop)
        self.i = 0
        self.last = None

    def loop( self,arg=None,st=0):
        if st: 
           print "Sleep for",st
           time.sleep(st)
        if hasattr(self.page,"callRemote") and hasattr(self,"datadirs"):
           datadir = self.datadirs[argmax([os.stat(datadir)[8] for datadir in self.datadirs])]
           if self.runner.fclass == "newfirm":
              ifiles = [g for g in glob(datadir+"/image?????.fits") if g not in self.ifiles]
           elif self.runner.fclass == "fourstar":
              ifiles = [g for g in glob(datadir+"/fsr_????_??_c?.fits") if g not in self.ifiles]
           if not ifiles:
              ifiles = self.ifiles
           ifiles.sort()
           if ifiles:
              ifile = ifiles[-1]
           if ifile != self.last:
              if self.runner.fclass == "newfirm":
                  F = FITS(ifile)
                  ra = F["ra"]
                  dec = F["dec"]
                  exptime = F["exptime"]
                  F.close()
                  js = []
                  lfile = ifile
                  for c in [1,2,3,4]:
                      b = os.path.basename(ifile).replace(".fits","c%d.jpg" % (c))
                      js += ["jpegs/"+b]
                  b = os.path.basename(ifile)
              elif self.runner.fclass == "fourstar":
                  F = FITS(ifile)
                  ra = F["ra"]
                  dec = F["dec"]
                  exptime = F["exptime"]
                  F.close()
                  js = []
                  idir = os.path.dirname(ifile)
                  lfile = idir+"/"+"_".join(os.path.basename(ifile).split("_")[:-1])+"_.fits"
                  for c in [1,2,3,4]:
                      b = os.path.basename(lfile).replace(".fits","c%d.jpg" % (c))
                      js += ["jpegs/"+b]
                  b = "_".join(os.path.basename(lfile).split("_")[:-1])

              if self.i:
                 self.callRemote('delText', u'ticker1', u'name')
                 self.callRemote('delText', u'ticker1', u'ra')
                 self.callRemote('delText', u'ticker1', u'dec')
                 self.callRemote('delText', u'ticker1', u'exptime')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%s' % (b), u'name')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%s' % (ra), u'ra')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%s' % (dec), u'dec')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%.1f' % (exptime), u'exptime')

              data = []
              for c in [1,2,3,4]:
                 ff = lfile.replace(".fits","c%d.fits" % (c))
                 b = os.path.basename(lfile).replace(".fits","c%d.jpg" % (c))
                 j1 = "jpegs/"+b
                 j2 = self.runner.basedir+"/jpegs/"+b
                 if os.path.exists(ff):
                    if c in self.chipim:
                       self.callRemote('delText', u'chip%d'%(c), u'chipim%d'%(c))
                       self.chipim.pop(c)
                    data.append([ff,c,j2])

#               self.wt = task.LoopingCall(self.WatchTicks,data)
#               self.wt = self.wt.start(0.1)

              send = lambda (ff,c,j1): self.TickOneStep2(ff,c,j1)
              Q = RunForkedQueue(send,data,MaxThreads=4,sleeptime=0.001)

              for c in [1,2,3,4]:
#                  if c in self.chipim:
#                     self.callRemote('delText', u'chip%d'%(c), u'chipim%d'%(c))
                  self.callRemote('addImage', u'chip%d'%(c), u'%s'%(j1), u'chipim%d'%(c))
                  self.callRemote('changeElement', u'style', u'width:90', u'chipim%d'%(c))
                  self.chipim[c] = 1

              self.last = ifile
              self.i += 1

        else:
           reactor.callLater(5,self.loop)

    def WatchTicks(self,data):
        if len([d for d in data if os.path.exists(d[2])]) == 4:
           reactor.callLater(2,self.loop)
           reactor.callLater(0,self.wt.stop)

    def TickOneStep2( self,ifile,c,j1):
           print self,ifile,c,j1
           F = FITS(ifile)
           ra = F["ra"]
           dec = F["dec"]
           exptime = F["exptime"]
           if self.runner.fclass == "newfirm":
              ncoadd = F["ncoadd"]
              if ncoadd == "N/A": ncoadd = 1
              F.goto(1+c)
           data = F.data()
           F.close()
           data = data[:2048,:2048]
           data = where(between(data,-1e24,1e24),data,0).astype(float32)
           easydir(self.runner.basedir+"/jpegs/")
           MakeThumbnail(data,mag=16,pct=40,s1=-10,s2=10,inv=1,q=85,j=j1)

           if self.runner.fclass == "newfirm":
               b = os.path.basename(ifile)
           elif self.runner.fclass == "fourstar":
               b = "_".join(os.path.basename(ifile).split("_")[:-1])

#           self.callRemote('addImage', u'chip%d'%(c), u'%s'%(j1), u'chipim%d'%(c))
#           self.callRemote('changeElement', u'style', u'width:90', u'chipim%d'%(c))

    def makeDocFactory( self):
        self.fireItUp()
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2["Frame Ticker"],
             T.table(style="border:thin solid black;")[
             T.thead[ T.tr[
                T.th(style="border:thin solid black;width:90")["Frame"],
                T.th(style="border:thin solid black;width:90")["RA"],
                T.th(style="border:thin solid black;width:90")["DEC"],
                T.th(style="border:thin solid black;width:30")["EXPTIME"],
                ] ],
             T.tbody[
                T.tr(name="ticker1")
                ],
                ],
             T.table(style="border:thin solid black;")[
             T.thead[ T.tr[
                T.th(style="border:thin solid black;width:75")["1"],
                T.th(style="border:thin solid black;width:75")["2"],
                T.th(style="border:thin solid black;width:75")["3"],
                T.th(style="border:thin solid black;width:75")["4"],
                ] ],
             T.tbody[ T.tr(name="ticker2")[
                  T.td[T.div(name=u"chip1")],
                  T.td[T.div(name=u"chip2")],
                  T.td[T.div(name=u"chip3")],
                  T.td[T.div(name=u"chip4")],
                  ] ],
                ],
                ])


class PipelineStatus(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def check( self):
        reactor.callInThread(self.loop)
#        self.clock = task.LoopingCall(self.loop)
#        self.clock.start(1)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        for n in range(10):
           if hasattr(self,"watch%d"%(n+1)):
              w = getattr(self,"watch%d"%(n+1))
              if w.running: w.stop()

    def loop( self):
        nwatch = [n for n in range(10) if hasattr(self,"watch%d"%(n)) and getattr(self,"watch%d"%(n)).running]
        if nwatch:
           print "Remaining watchers",nwatch
           reactor.callLater(1,self.loop)
           return
        d = []
        if hasattr(self.page,"callRemote") and self.runner.going != self.status:
           if self.runner.going:
              status = "Running"
              color = "green"
           else:
              status = "Paused"
              color = "red"
           d.append(self.callRemote('changeElement', u'style', u"color:%s"%(color), u"pstatus"))
           d.append(self.callRemote('delText', u"pstatus", u"status"))
           d.append(self.callRemote('addText', u"span", u"pstatus", u"%s"%(status), u"status"))
        self.status = self.runner.going
        if d:
           self.loopcount3 = 0
           self.watch3 = task.LoopingCall(self.notify,d,self.loop,3,1.0)
           self.watch3.start(1.0)

    def notify( self,dall,f,n=1,s=0.0,to=60):
        dall = [d for d in dall if d]
        if not hasattr(self,"looptimes"):
           self.looptimes = {}
        if n in self.looptimes: self.looptimes[n] += 1.0
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n))
#        if rem == 1:
#           print [d for d in d if not hasattr(d,"result")]
        if rem and self.looptimes[n]<to: return
        w = eval("self.watch%d"%(n))
        w.stop()
        self.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)

    def select( self, value=None):
        self.runner.toggleGoing()
        self.loop()
    athena.expose(select)

    def makeDocFactory( self):
#        self.check()
        self.status = self.runner.going
        click = u'Nevow.Athena.Widget.get(this).selected(\'switch\')'
        if self.runner.going:
           status = "Running"
           color = "green"
        else:
           status = "Paused"
           color = "red"
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2[
             T.span["Pipeline Status:"],
             T.span(style=u"color:%s"%(color),name=u"pstatus",onclick=click)[
                T.span(onmouseover=mouseover,onmouseout=mouseout,name=u"status")[status]
                ]
                ],
                ])


class TargetsTable(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.qstats = {}
        self.statuslist = {}
        self.listing = {}

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def check( self):
#        self.current = None
        self.showing = []
#        reactor.callInThread(self.loop)
#        self.clock = task.LoopingCall(self.loop)
#        self.clock.start(1)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        for n in range(10):
           if hasattr(self,"watch%d"%(n+1)):
              w = getattr(self,"watch%d"%(n+1))
              if w.running: w.stop()

    def getstatus( self, targ):
         itarg = self.runner.targs.index(targ)+1
         name = u"targ(%d)" % (itarg)
         name1 = u"targ(%d)1" % (itarg)
         name2 = u"targ(%d)2" % (itarg)
         name3 = u"targ(%d)3" % (itarg)
         cname1 = u"ctarg(%d)1" % (itarg)
         cname2 = u"ctarg(%d)2" % (itarg)
         cname3 = u"ctarg(%d)3" % (itarg)
         bname1 = u"btarg(%d)1" % (itarg)
         bname2 = u"btarg(%d)2" % (itarg)
         bname3 = u"btarg(%d)3" % (itarg)
         self.runner.Data1[targ].Already(first=True,second=False)
         self.runner.Data2[targ].Already(first=False,second=True)
         done1 = self.runner.Data1[targ].donefirst
         done2 = self.runner.Data2[targ].donesecond
         err1 = self.runner.Data1[targ].errfirst
         err2 = self.runner.Data2[targ].errsecond
         ongoing1 = self.runner.Data1[targ].ongoing
         ongoing2 = self.runner.Data2[targ].ongoing
         bstyle1 = u"width:300;font-weight:normal"
         bstyle2 = u"width:20;font-weight:normal"
         bstyle3 = u"width:20;font-weight:normal"
         status = self.runner.status[targ]
         if err1 or err2:
            bstyle1 += u";color:red"
            bstyle2 += u";color:red"
            bstyle3 += u";color:red"
            status = "Error!"
         elif done1 and not done2:
            bstyle1 += u";color:orange"
            bstyle2 += u";color:orange"
            bstyle3 += u";color:orange"
            status = self.runner.status[targ]
            status = "Done!"
         elif done1 and done2:
            bstyle1 += u";color:green"
            bstyle2 += u";color:green"
            bstyle3 += u";color:green"
            status = self.runner.status[targ]
            status = "Done!"
         else:
            bstyle1 += u";color:blue"
            bstyle2 += u";color:blue"
            bstyle3 += u";color:blue"
            status = self.runner.status[targ]
         if ongoing1 or ongoing2:
            bstyle1 += u";font-weight:bold"
            status = "Ongoing!"
         else:
            bstyle1 += u";font-weight:normal"
         content1 = u"%s" % (targ)
         content2 = u"%d" % (len(self.runner.groups[targ]))
         content3 = u"%s" % (status)
         click1 = u'Nevow.Athena.Widget.get(this).selected(\'%s\')' % (targ)
         click3 = u'Nevow.Athena.Widget.get(this).selected2(\'%s\')' % (targ)
         return status,done1,done2,ongoing1,ongoing2,name,cname1,cname2,cname3,\
                bstyle1,bstyle2,bstyle3,click1,click3,content1,content2,content3

    def include( self, targ):
        if self.gtype == "all":
           return True
        elif self.gtype == "targets":
           return (hasattr(self.runner,"Data1") and not self.runner.Data1[targ].flat) or \
                (hasattr(self.runner,"Data2") and not self.runner.Data2[targ].flat)
        elif self.gtype == "calibs":
           return (hasattr(self.runner,"Data1") and self.runner.Data1[targ].flat) or \
                (hasattr(self.runner,"Data2") and self.runner.Data2[targ].flat)

    def loop( self):
        nwatch = [n for n in range(10) if hasattr(self,"watch%d"%(n)) and getattr(self,"watch%d"%(n)).running]
        if nwatch:
           print "Remaining watchers",nwatch
           reactor.callLater(1,self.loop)
           return
        if hasattr(self.page,"callRemote"):
           dall = []
           showtargs = [targ for targ in self.runner.targs if targ]
           for targ in showtargs:
               inc = self.include(targ)
               if inc and (targ not in self.targets or targ not in self.listing):
                  self.targets.append(targ)
                  status,done1,done2,ongoing1,ongoing2,name,cname1,cname2,cname3,\
                      bstyle1,bstyle2,bstyle3,click1,click3,content1,content2,content3 = \
                          self.getstatus(targ)
                  print "Adding status and styles for",targ
                  self.targets.append(targ)
                  self.statuslist[targ] = status
                  self.qstats[targ] = (done1,done2,ongoing1,ongoing2)
                  d1 = self.callRemote('addElement', u'tr', u'', u'targetTableBody', u'', name)
                  d2 = self.callRemote('addElement', u'td', u'', name, u'', cname1+"d")
                  d3 = self.callRemote('addElement', u'td', u'', name, u'', cname2+"d")
                  d4 = self.callRemote('addElement', u'td', u'', name, u'', cname3+"d")
                  d5 = self.callRemote('addElement', u'div', bstyle1, cname1+"d", u'%s' % (targ), cname1)
                  d6 = self.callRemote('addElement', u'div', bstyle2, cname2+"d", u'%d' % (len(self.runner.groups[targ])), cname2)
                  d7 = self.callRemote('addElement', u'div', bstyle3, cname3+"d", u'%s' % (status), cname3)
                  d8 = self.callRemote('changeElement', u'onmouseover', mouseover, name)
                  d9 = self.callRemote('changeElement', u'onmouseout', mouseout, name)
                  dA = self.callRemote('changeElement', u'onclick', click1, cname1)
                  dB = self.callRemote('changeElement', u'onclick', click3, cname3)
                  dall.extend([d1,d2,d3,d4,d5,d6,d7,d8,d9,dA,dB])
                  self.listing[targ] = [name,cname1,cname2,cname3]
               elif not inc and (targ in self.targets and targ in self.listing):
                  name,cname1,cname2,cname3 = self.listing[targ]
                  print "Deleting",targ
                  d1 = self.callRemote('delText',cname1+"d",cname1)
                  d2 = self.callRemote('delText',cname2+"d",cname2)
                  d3 = self.callRemote('delText',cname3+"d",cname3)
                  d4 = self.callRemote('delText',name,cname1+"d")
                  d5 = self.callRemote('delText',name,cname2+"d")
                  d6 = self.callRemote('delText',name,cname3+"d")
                  d7 = self.callRemote('delText', u'targetTableBody', name)
                  dall.extend([d1,d2,d3,d4,d5,d6,d7])
                  self.listing.pop(targ)
               elif inc:
                  status,done1,done2,ongoing1,ongoing2,name,cname1,cname2,cname3,\
                      bstyle1,bstyle2,bstyle3,click1,click3,content1,content2,content3 = \
                          self.getstatus(targ)
#                  print targ,status,self.statuslist.get(targ,None)
                  if targ not in self.qstats or \
                        self.qstats[targ] != (done1,done2,ongoing1,ongoing2) or \
                        self.listing[targ] != [name,cname1,cname2,cname3] or \
                             self.statuslist[targ] != status:
                     self.qstats[targ] = (done1,done2,ongoing1,ongoing2)
                     print "Changing status and styles for",targ
                     self.statuslist[targ] = status
                     if targ == self.page.targetWatcher.target:
                       bstyle1 = bstyle1 + u";text-decoration:underline"
                       bstyle2 = bstyle2 + u";text-decoration:underline"
                       bstyle3 = bstyle3 + u";text-decoration:underline"
                     content1 = u"%s" % (targ)
                     content2 = u"%d" % (len(self.runner.groups[targ]))
                     content3 = u"%s" % (status)
                     d1 = self.callRemote('delText',cname3+"d", cname3)
                     d2 = self.callRemote('addElement', u'div', bstyle3, cname3+"d", u'%s' % (status), cname3)
                     d3 = self.callRemote('changeElement', u'style', bstyle1, cname1)
                     d4 = self.callRemote('changeElement', u'style', bstyle2, cname2)
                     d5 = self.callRemote('changeElement', u'style', bstyle3, cname3)
                     dall.extend([d1,d2,d3,d4,d5])
                     self.listing[targ] = [name,cname1,cname2,cname3]
           if len(dall):
              self.loopcount0 = 0
              self.watch0 = task.LoopingCall(self.notify,dall,lambda : self.loop(),0,1.0)
              self.watch0.start(1.0)
           else:
              reactor.callLater(1.0,self.loop)

    def notify( self,dall,f,n=1,s=0.0,to=60,clear=True):
        dall = [d for d in dall if d]
        if not hasattr(self,"looptimes"):
           self.looptimes = {}
        if n in self.looptimes: self.looptimes[n] += 1.0
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n))
#        if rem == 1:
#           print [d for d in d if not hasattr(d,"result")]
        if rem and self.looptimes[n]<to: return
        w = eval("self.watch%d"%(n))
        w.stop()
        if clear: self.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)

    def select( self, value=None):
        if value:
           otarg = self.page.targetWatcher.target
           dall = []
           print value,otarg
           if otarg in self.runner.targs:
              itarg = self.runner.targs.index(otarg)+1
              name = u"targ(%d)" % (itarg)
              name1 = u"ctarg(%d)1" % (itarg)
              dall.append(self.callRemote('changeElement',u"style",u"color:black;text-decoration:normal",name))
           itarg = self.runner.targs.index(value)+1
           name = u"targ(%d)" % (itarg)
           name1 = u"ctarg(%d)1" % (itarg)
           dall.append(self.callRemote('changeElement',u"style",u"color:black;text-decoration:underline",name))
           dall.extend(self.page.targetWatcher.fireUp(value))
           self.loopcount1 = 0
           self.watch1 = task.LoopingCall(self.notify,dall,None,1)
           self.watch1.start(1.0)
    athena.expose(select)

    def moment( self, dummy="", stuff=""):
        print dummy,stuff

    def select2( self, value=None):
        if value and value in self.runner.status:
           targs = [value]
        elif value == "all":
           targs = [targ for targ in self.runner.status.keys() if targ]
        else:
           targs = []
        for targ in targs:
           if targ in self.listing:
              itarg = self.runner.targs.index(targ)+1
              done1 = self.runner.Data1[targ].donefirst
              done2 = self.runner.Data2[targ].donesecond
              err1 = self.runner.Data1[targ].errfirst
              err2 = self.runner.Data2[targ].errsecond
              ongoing1 = self.runner.Data1[targ].ongoing
              ongoing2 = self.runner.Data2[targ].ongoing
              if err1 or err2:
                 status = "Error!"
              elif done1 and not done2:
                 status = "Done!"
              elif done1 and done2:
                 status = "Done!"
              else:
                 status = not self.runner.status[targ]
                 self.runner.status[targ] = not self.runner.status[targ]
              print "Status!",targ, status, ongoing1, ongoing2
              if ongoing1 or ongoing2:
                 status = "Ongoing!"
              print "Status!",targ, status, ongoing1, ongoing2
#              self.statuslist[targ] = status
              d = self.callRemote("cursor_clear")
#        reactor.callLater(0.5,self.loop)
    athena.expose(select2)

    def select3( self, value=None):
        print self.gtype, value
        if self.gtype == "all":
           self.gtype = "targets"
           content = u" (Targets)"
        elif self.gtype == "targets":
           self.gtype = "calibs"
           content = u" (Calibrations)"
        elif self.gtype == "calibs":
           self.gtype = "all"
           content = u" (Calibrations+Targets)"
        d1 = self.callRemote('delText',u"groups",u"gtype")
        d2 = self.callRemote('addText',u"span",u"groups",content,u"gtype")
        dall = [d1,d2]
        self.loopcount2 = 0
        self.watch2 = task.LoopingCall(self.notify,dall,None,2,clear=False)
        self.watch2.start(1.0)
    athena.expose(select3)

    def makeDocFactory( self):
        self.check()
        self.targets = []
        reactor.callLater(0.5,self.loop)
        self.gtype = "all"
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2[T.div(onclick='Nevow.Athena.Widget.get(this).selected3(\'gtype\')')
                [
                  T.span(onmouseover=mouseover,onmouseout=mouseout,name=u"gdirs")[
                      T.span(name=u"gdir")["Exposure Groups:"],
                        ],
                  T.span(onmouseover=mouseover,onmouseout=mouseout,name=u"groups")[
                      T.span(name=u"gtype")[" (Calibrations+Targets)"]
                      ]
                      ]
                ],
             T.div(style=u'width:400;height:480;overflow:auto')[
             T.table(style=u"border:thin solid black",name="targetTable")[
             T.thead[ T.tr[
                 T.th(style=u"border:thin solid black;")["Grouping"],
                 T.th(style=u"border:thin solid black;")["N"],
                 T.th(onclick='Nevow.Athena.Widget.get(this).selected2(\'%s\')' % ('all'),\
                        style=u"border:thin solid black;")[
                    T.div(onmouseover=mouseover,onmouseout=mouseout)[T.div["Run?"]]
                        ],
                 ] ],
             T.tbody(name='targetTableBody')
                ]]])

class TargetWatcher(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        self.target = "No Target Selected"
        self.showP = 0
        self.showS = 0
        self.schunk = 2

        self.npasses = []
        self.nstages = {0:[1,]}
        self.tstages = {0:["N/A"]}

        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.images = {}

    def cleanUp( self):
        d = []
        self.callRemote('cursor_wait')
        for p in self.npasses:
            for s in self.nstages[p]:
                print p,s
                d.append(self.callRemote('delText',u"sel%d"%(100*p+s), u"sel%dt"%(100*p+s)))
        for p in self.npasses:
            for s in self.nstages[p]:
                d.append(self.callRemote('delText',u"thumbs", u"sel%d"%(100*p+s)))
            d.append(self.callRemote('delText',u'thumbs',u'selbr%d'%(p)))
        return d

    def loop( self):
        nwatch = [n for n in range(10) if hasattr(self,"watch%d"%(n)) and getattr(self,"watch%d"%(n)).running]
        if nwatch:
           print "Remaining watchers",nwatch
           reactor.callLater(1,self.loop)
           return
        style = u"color:orange"
        d = []
        for p in self.npasses:
            for s in self.nstages[p]:
                if "orange" in self.sdict[p][s] or (p == self.showP and s == self.showS):
                   tdir = self.gettdir(self.target,p,s)
                   timages = self.thumbpics(tdir)
                   if timages: style = u"color:green;font-weight:normal;"
                   else: style = u"color:orange;font-weight:normal;"
                   self.sdict[p][s] = style
                   if p == self.showP and s == self.showS: style = style + u"text-decoration:underline"
                   else: style = style + u"text-decoration:normal"
                   d.append(self.callRemote('changeElement',u'style',style,u"sel%dt"%(100*p+s)))
        if d:
           self.loopcount2 = 0
           self.watch2 = task.LoopingCall(self.notify,d,self.loop,2,1.0)
           self.watch2.start(1.0)
        else:
           reactor.callLater(1,self.loop)

    def addUp1( self, target, p, s):
        ptitle = lambda p,s: u"(%d) %s" % (p,self.tstages[p][s-1])
        d = []
        d1 = self.callRemote('addElement',u"div",u'',u"thumbs",u'',u"sel%d"%(100*p+s))
        d2 = self.callRemote('addText',u"div",u"sel%d"%(100*p+s),ptitle(p,s),u"sel%dt"%(100*p+s))
        d.extend([d1,d2])
        return d

    def changeUp1( self, target, p, s):
        print "in change up1",target,p,s
        tdir = self.gettdir(target,p,s)
        timages = self.thumbpics(tdir)
        if timages: style = u"color:green;font-weight:normal;"
        else: style = u"color:orange;font-weight:normal;"
        self.sdict[p][s] = style
        if p == self.showP and s == self.showS: style = style + u"text-decoration:underline"
        else: style = style + u"text-decoration:normal"
#        print p,s,style

        d = []
        d.append( self.callRemote('changeElement',u"style",style,u"sel%dt"%(100*p+s)))
        d.append( self.callRemote('changeElement',u'onmouseover',mouseover,u"sel%dt"%(100*p+s)))
        d.append( self.callRemote('changeElement',u'onmouseout',mouseout,u"sel%dt"%(100*p+s)))
        d.append( self.callRemote('changeElement',u"onclick",u'Nevow.Athena.Widget.get(this).selected(\'%d\')' % (100*p+s),u"sel%dt"%(100*p+s)))
        return d

    def addUp( self, target):
        d = []
        for p in self.npasses:
            for s in self.nstages[p]:
                d.extend(self.addUp1(target, p, s))
            d.append(self.callRemote('addElement',u'br',u'',u'thumbs',u'',u'selbr%d'%(p)))
        return d

    def changeUp( self, target):
        d = []
        self.sdict = {}
        for p in self.npasses:
            self.sdict[p] = {}
            for s in self.nstages[p]:
                d.extend(self.changeUp1(target, p, s))
        return d

    def oldchangeUp( self, target,d1):
        ptitle = lambda p,s: u"(%d) %s" % (p,self.tstages[p][s-1])
        self.sdict = {}
        for p in self.npasses:
            self.sdict[p] = {}
            for s in self.nstages[p]:
                tdir = self.gettdir(target,p,s)
                timages = self.thumbpics(tdir)
                if timages: style = u"color:green;font-weight:normal;"
                else: style = u"color:orange;font-weight:normal;"
                self.sdict[p][s] = style
                if p == self.showP and s == self.showS: style = style + u"text-decoration:underline"
                else: style = style + u"text-decoration:normal"
                d1.addCallback(callRemote,self,'changeElement',u"style",style,u"sel%dt"%(100*p+s))
                d1.addCallback(callRemote,self,'changeElement',u'onmouseover',mouseover,u"sel%d"%(100*p+s))
                d1.addCallback(callRemote,self,'changeElement',u'onmouseout',mouseout,u"sel%d"%(100*p+s))
                d1.addCallback(callRemote,self,'changeElement',u"onclick",u'Nevow.Athena.Widget.get(this).selected(\'%d\')' % (100*p+s),u"sel%d"%(100*p+s))

    def moment( self, dummy="", stuff=""):
        print dummy,stuff

    def fireUp2( self, target):
        dall = []
        dall.append(self.callRemote('delText', u"selheader", u"seltarget"))
        tstring = u"Target: %s" % (target)
        dall.append(self.callRemote('addText', u"div", u"selheader", tstring, u"seltarget"))
        dall.extend(self.cleanUp())
        self.setUp(target=target)
        dall.extend(self.addUp(target))
        dall.extend(self.changeUp(target))
        self.setTarg(target=target)
        self.loopcount4 = 0
        self.watch4 = task.LoopingCall(self.notify,dall,self.loop,4)
        self.watch4.start(1.0)
        return dall

    def fireUp( self, target):
        print "fireUp!",target
        dall = []
        if self.target == target:
           dall.append(self.callRemote('cursor_clear'))
        else:
           if self.images:
              d2 = self.clearsnaps()
              self.loopcount5 = 0
              self.watch5 = task.LoopingCall(self.notify,d2,lambda: self.fireUp2(target),5)
              self.watch5.start(1.0)
           else:
              dall.extend(self.fireUp2(target))
        return dall

    def notify( self,dall,f,n=1,s=0.0,to=60):
        dall = [d for d in dall if d]
        if not hasattr(self,"looptimes"):
           self.looptimes = {}
        if n in self.looptimes: self.looptimes[n] += 1.0
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n))
#        if rem == 1:
#           print [d for d in d if not hasattr(d,"result")]
        if rem and self.looptimes[n]<to: return
        w = eval("self.watch%d"%(n))
        w.stop()
        self.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)


    def setTarg( self, *args, **kws):
        target = kws["target"]
        if target:
           print "Setting target to:",target
           self.target = target

    def setUp( self, *args, **kws):
        target = kws["target"]
        print "In setUp"
        if self.runner.Data1[target].dark:
           self.npasses = [1]
           self.nstages = {1:[1,2,3]}
           self.tstages = {1:["Calibrated Frames","Combined Frames","Darks"]}
        elif self.runner.Data1[target].flat:
           self.npasses = [1]
           self.nstages = {1:[1,2,3,4]}
           self.tstages = {1:["Calibrated Frames","Combined Frames","Flats","Flat Masks"]}
        else:
           self.npasses = [1,2]
           self.nstages = {1:[1,2,3,4,5,6],2:[1,2,3,4,5,6,7,8,9]}
           self.tstages = {1:["Calibrated Frames","Simple Sky Frames","Sky Subtracted Frames","Shift Plots",\
                                   "Stacked Frames","Mosaicked Frames"],
                           2:["Calibrated Frames","Interpolated Frames","Wavelet Transforms",\
                                   "Model Transforms","Model Sky Frames","Sky Subtracted Frames",\
                                           "Shift Plots","Stacked Frames","Mosaicked Frames"]}
        self.showP = 0
        self.showS = 0

    def select( self, value=None):
        showP = int(value)/100
        showS = int(value)%100

        if self.target in self.runner.targs and value:
          name = u"%s" % (self.target.replace(" ",""))
          if hasattr(self,"callRemote"):
             dall = []
#             style = u"text-decoration:underline"
#             item = u"sel%d"%(100*showP+showS)
#             dal.append(self.callRemote('changeElement',u"style",style,item))
             tdir = self.gettdir(self.target,showP,showS)
             timages = self.thumbpics(tdir)
             if timages: style = u"font-weight:normal;color:green;text-decoration:underline"
             else: style = u"font-weight:normal;color:orange;text-decoration:underline"
             item = u"sel%d"%(100*showP+showS)
             dall.append(self.callRemote('changeElement',u"style",style,item))

             dall.append(self.callRemote('delText',u"hproducts",u"products"))
             text = u"Products for %s (Pass %d, %s)" % (name,showP,self.tstages[showP][showS-1])
             dall.append(self.callRemote('addText',u'div',u"hproducts",text,u"products"))
             if self.showP and self.showS:
                tdir = self.gettdir(self.target,self.showP,self.showS)
                timages = self.thumbpics(tdir)
                if timages: style = u"font-weight:normal;color:green;"
                else: style = u"font-weight:normal;color:orange;"
                if showP == self.showP and showS == self.showS:
                   style = style + u"text-decoration:underline"
                else:
                   style = style + u"text-decoration:normal"
                item = u"sel%d"%(100*self.showP+self.showS)
                dall.append(self.callRemote('changeElement',u"style",style,item))
             self.showP = showP
             self.showS = showS
             dall.extend(self.addThumbs())
             self.loopcount9 = 0
             self.watch9 = task.LoopingCall(self.notify,dall,None,9)
             self.watch9.start(0.5)
        else:
          self.showP = 0
          self.showS = 0
    athena.expose(select)

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        self.clock.stop()
        for n in range(10):
           if hasattr(self,"watch%d"%(n+1)):
              w = getattr(self,"watch%d"%(n+1))
              if w.running: w.stop()

    def fireUpProducts( self):
        d = []
        if hasattr(self.page,"callRemote"):
           if self.images:
              d2 = self.clearsnaps()
              self.loopcount8 = 0
              self.watch8 = task.LoopingCall(self.notify,d2,self.snapshot,8)
              self.watch8.start(0.5)
           else:
              self.snapshot()
        return d

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        for n in range(10):
           if hasattr(self,"watch%d"%(n+1)):
              w = getattr(self,"watch%d"%(n+1))
              if w.running: w.stop()

    def clearsnap( self,pics,j):
        d = []
        for pic in pics[j*self.schunk:(j+1)*self.schunk]:
            lpic = pic.replace("."+pic.split(".")[-1],"link")
            finfo = self.images[pic]
            if 0:
               d.append(self.callRemote( 'delText',lpic,pic))
               for k,ftext in enumerate(finfo):
                   if ftext:
                      fwords = ftext.split()
                      nwords = len(fwords)
                      for l in range(nwords):
                          d.append(self.callRemote('delText', u'thumb%sinforow%d' % (pic,k+1), u'thumb%sinfo%d' % (pic,k*nwords+l)))

            d.append(self.callRemote( 'removeRows',u"thumb%sinfotbl" % (pic)))
            d.append(self.callRemote( 'delText',u"thumb%sinfo" %(pic),u"thumb%sinfotbl" % (pic)))
            d.append(self.callRemote( 'delText',u"thumb%s" %(pic),lpic))
            d.append(self.callRemote( 'delText',u"thumb%srow1" %(pic),u"thumb%s" % (pic)))
            d.append(self.callRemote( 'delText',u"thumb%srow2" %(pic),u"thumb%sinfo" % (pic)))
            d.append(self.callRemote( 'removeRows',u"thumb%stbl" % (pic)))
            d.append(self.callRemote( 'delText',u"thumb%sitem" %(pic),u"thumb%stbl" % (pic)))
            d.append(self.callRemote( 'delText',u"thumbrow%d" %(j+1),u"thumb%sitem" % (pic)))
            self.images.pop(pic)
        return d

    def clearsnaps( self):
        print "in clearsnaps",self.images
        d = []
        if self.images:
           print "removing previous images"
#           d.append(self.callRemote('delText',u"hproducts",u"products"))
#           d.append(self.callRemote('addText',u'div',u"hproducts",u"Clearing Snapshots",u"products"))
           pics = sorted(self.images.keys())
           nchunk = (len(pics)-1)/self.schunk + 1
           for j in range(nchunk):
               d.extend( self.clearsnap(pics,j))
           d.append( self.callRemote('removeRows',u"tablethumbs"))
           d.append( self.callRemote( 'delText',u"stagethumbs",u"tablethumbs"))
        return d

    def snapshot( self):
#        print "in snapshot"
        d = []
        if hasattr(self.page,"callRemote") and self.showS and self.showP:
#           d.append(self.callRemote('delText',u"hproducts",u"products"))
           name = u"%s" % (self.target.replace(" ",""))
           text = u"Products for %s (Pass %d, %s)" % (name,self.showP,self.tstages[self.showP][self.showS-1])
#           d.append(self.callRemote('addText',u'div',u"hproducts",text,u"products"))
           tdir = self.gettdir(self.target,self.showP,self.showS)
           if u"Stack" in tdir: self.schunk = 3
           elif u"FirstSky" in tdir: self.schunk = 2
           elif u"shift" in tdir: self.schunk = 3
           elif u"Flat" in tdir: self.schunk = 2
           elif u"Dark" in os.path.basename(tdir): self.schunk = 2
           elif u"dark" in tdir and u"sky" not in tdir and u"mode" not in tdir: self.schunk = 2
           elif u"mask" in tdir: self.schunk = 2
           else: self.schunk = 4
           d.extend(self.thumbs("/"+tdir))
           self.loopcount7 = 0
           self.watch7 = task.LoopingCall(self.notify,d,None,7)
           self.watch7.start(0.5)

        return d

    def gettdir( self,target,showP,showS):
        if self.runner.fclass == "newfirm": base = "image"
        elif self.runner.fclass == "fourstar": base = "fsr_"
        if showP == 1:
           if showS == 1: tdir = u"%s/mode%d/%s" % (target,showP,base)
           elif showS == 2:
               if self.runner.Data1[target].dark:
                  tdir = u"%s/sky%d/Dark" % (target,showP)
               elif self.runner.Data1[target].flat:
                  tdir = u"%s/sky%d/Flat" % (target,showP)
               else:
                  tdir = u"%s/sky%d/FirstSky" % (target,showP)
           elif showS == 3:
               if self.runner.Data1[target].dark:
                  tdir = u"%s/dark" % (target)
               elif self.runner.Data1[target].flat:
                  tdir = u"%s/Flat" % (target)
               else:
                  tdir = u"%s/sky%d/%s" % (target,showP,base)
           elif showS == 4:
               if self.runner.Data1[target].flat:
                  tdir = u"%s/mask" % (target)
               else:
                  tdir = u"%s/shift%d/%s" % (target,showP,base)
           elif showS == 5: tdir = u"%s/Stack%d_" % (target,showP)
           elif showS == 6: tdir = u"%s/Mosaic%d" % (target,showP)
        elif showP == 2:
           if showS == 1: tdir = u"%s/mode%d/%s" % (target,showP,base)
           elif showS == 2: tdir = u"%s/back%d/%s" % (target,showP,base)
           elif showS == 3: tdir = u"%s/mall%d/%s" % (target,showP,base)
           elif showS == 4: tdir = u"%s/smow%d/%s" % (target,showP,base)
           elif showS == 5: tdir = u"%s/mbg%d/%s" % (target,showP,base)
           elif showS == 6: tdir = u"%s/sky%d/%s" % (target,showP,base)
           elif showS == 7: tdir = u"%s/shift%d/%s" % (target,showP,base)
           elif showS == 8: tdir = u"%s/Stack%d_" % (target,showP)
           elif showS == 9: tdir = u"%s/Mosaic%d" % (target,showP)
        return tdir

    def waitFor(self,d):
        while not d.called:
            print "Waiting on",d,d.called
            time.sleep(1)
        print "Called:",d,d.called

    def dothumb(self,r,j,pic,tdir):
        suff = u"%s" % (pic.split(".")[-1])
        lpic = u"%s" % (pic.replace("."+suff,"link"))
        upic = u"%s" % (tdir+pic.replace("mag16."+suff,"mag01."+suff).replace("mag08."+suff,"mag01."+suff))
        fpic = u"%s" % (pic.replace("mag16","").replace("mag08","").replace("."+suff,""))
        finfo = [
                fpic,
                ""
                ]

        if "shift%d" % (self.showP) in tdir:
           frame = glob(u"%s/%s/sky%d/%sc?.fits" % (self.runner.basedir,self.target,self.showP,fpic))
           if frame:
              F = FITS(frame[0])
              xdiff1 = F["xdiff1"]
              ydiff1 = F["ydiff1"]
              mdiff1 = F["mdiff1"]
              xscat1 = F["xscat1"]
              yscat1 = F["yscat1"]
              mscat1 = F["mscat1"]
              lmatch1 = F["lmatch1"]
              F.close()
              finfo = [
                      u"%s %6.2f %6.2f %6.2f" % (fpic,xdiff1,ydiff1,mdiff1),
                      u"%d %6.2f %6.2f %6.2f" % (lmatch1,xscat1,yscat1,mscat1),
                      ]
        elif "mode%d" % (self.showP) in tdir:
           frame = glob(u"%s/%s/mode%d/%s.fits" % (self.runner.basedir,self.target,self.showP,fpic))
           if frame:
              F = FITS(frame[0])
              medval = F["median"]
              scat = F["scat"]
              F.close()
              finfo = [
                      u"%s" % (fpic),
                      u"median=%8.2f" % (medval),
                      u"sigma=%8.2f" % (scat),
                      ]
        elif "sky%d" % (self.showP) in tdir and "Flat" not in fpic and "Dark" not in fpic:
           frame = glob(u"%s/%s/sky%d/%s.fits" % (self.runner.basedir,self.target,self.showP,fpic))
           if frame:
              F = FITS(frame[0])
              medval = F["median"]
              anoise = F["anoise"]
              F.close()
              enoise = getsqrt(8*medval + 25**2)/8.
              if "FirstSky" in fpic or "SecondSky" in fpic:
                 finfo = [
                         u"%s" % (fpic),
                         ]
              else:
                 finfo = [
                         u"%s %6.2f %6.2f" % (fpic,anoise,enoise),
                         ]
        else:
           finfo = [
                   fpic,
                   ""
                   ]
        d = []

        d.append(self.callRemote('addElement', u'td', u"border:thin solid black;", u'thumbrow%d' % (r+1), u'', u'thumb%sitem' % (pic)))
        d.append(self.callRemote('addElement', u'table', u"", u'thumb%sitem' % (pic), u'', u'thumb%stbl' % (pic)))
        d.append(self.callRemote('addElement', u'tr', u'border: none', u'thumb%stbl' % (pic), u'', u'thumb%srow1' % (pic)))
        d.append(self.callRemote('addElement', u'td', u"", u'thumb%srow1' % (pic), u'', u'thumb%s' % (pic)))
        d.append(self.callRemote('addElement', u'tr', u'border: none', u'thumb%stbl' % (pic), u'', u'thumb%srow2' % (pic)))
        d.append(self.callRemote('addElement', u'td', u"", u'thumb%srow2' % (pic), u'', u'thumb%sinfo' % (pic)))
        d.append(self.callRemote('addElement', u'table', u"", u'thumb%sinfo' % (pic), u'', u'thumb%sinfotbl' % (pic)))

        ninfo = len(finfo)
        for j,ftext in enumerate(finfo):
            if ftext:
               d.append(self.callRemote('addElement', u'tr', u'border: none', u'thumb%sinfotbl' % (pic), u'', u'thumb%sinforow%d' % (pic,j+1)))
               fwords = ftext.split()
               nwords = len(fwords)
               for i in range(nwords):
                   d.append(self.callRemote('addElement', u'td', u"", u'thumb%sinforow%d' % (pic,j+1), u"%s" % (fwords[i]), u'thumb%sinfo%d' % (pic,j*nwords+i)))

        d.append(self.callRemote( 'addElement', u'a', u'', u'thumb%s' % (pic), u'', lpic))

        d.append(self.callRemote( 'changeElement',u'onmouseover',mouseover,u'thumb%sitem' % (pic)))
        d.append(self.callRemote( 'changeElement',u'onmouseout',mouseout,u'thumb%sitem' % (pic)))
        d.append(self.callRemote( 'changeElement', u'href', upic, lpic))
        d.append(self.callRemote( 'changeElement', u'target', u"_blank", lpic))
        d.append(self.callRemote( 'addImage', lpic, tdir+pic, pic))
        if "Mosaic" in pic:
           d8 = self.callRemote( 'changeElement', u'style', u'width:512', pic)
        elif "Stack" in pic:
           d8 = self.callRemote( 'changeElement', u'style', u'width:256', pic)
        elif "FirstSky" in pic:
           d8 = self.callRemote( 'changeElement', u'style', u'width:256', pic)
        elif "Flat" in pic:
           d8 = self.callRemote( 'changeElement', u'style', u'width:256', pic)
        elif "Dark" in os.path.basename(pic) or "dark" in os.path.basename(pic):
           d8 = self.callRemote( 'changeElement', u'style', u'width:256', pic)
        elif "mag16" in pic:
           d8 = self.callRemote( 'changeElement', u'style', u'width:128', pic)
        else:
           d8 = self.callRemote( 'changeElement', u'style', u'width:256', pic)
        d.append(d8)
        return finfo,d

    def thumbpics(self,tdir):
        items1 = glob(self.runner.basedir+"/"+tdir+"*.jpg")
        items2 = glob(self.runner.basedir+"/"+tdir+"*.png")
        bdir = "/".join(tdir.split("/")[:-1]) + "/"
        items = sorted(items1 + items2)
        grab = lambda item: item[-9:] in ["mag16.jpg","mag16.png","mag08.jpg","mag08.png"] or item.split(".")[-1] in ["png"]
        pics = [u"%s" % (os.path.basename(item)) for item in items if grab(item)]
        return pics

    def thumbs(self,tdir):
        d = []
        pics = self.thumbpics(tdir)
        nchunk = (len(pics)-1)/self.schunk + 1
        bdir = "/".join(tdir.split("/")[:-1]) + "/"
        d.append(self.callRemote('addElement', u'table', u"border: none", u'stagethumbs', u'', u'tablethumbs'))
        for j in range(nchunk):
            d.append(self.callRemote('addElement', u'tr', u'border: none', u'tablethumbs', u'', u'thumbrow%d' % (j+1)))
            for p,pic in enumerate(pics[j*self.schunk:(j+1)*self.schunk]):
                finfo,d2 = self.dothumb(j,p+j*self.schunk,pic,bdir)
                d.extend(d2)
                self.images[pic] = finfo
        return d

    def addThumbs(self):
        d = []
        d.extend(self.fireUpProducts())
        return d

    def table( self):
        name = u"%s" % (self.target.replace(" ",""))
        ptitle = lambda p,s: u"(%d) %s" % (p,self.tstages[p][s-1])
        p = T.div(style=u'position:relative;',name=u"productTable")[
           T.table(style="border:thin solid black;")[
           T.thead[T.tr[
              T.th(style="width:180;border:thin solid black;")["(Pass) Stage"],
              T.th(style="width:800;border:thin solid black;",name=u"hproducts")[T.div(name=u"products")["Select a Stage"]],
              ]],
           T.tbody[T.tr[
              T.td(style="border:thin solid black;",name=u"thumbs")
                [
                   (((s < 999 and T.div(name=u"sel%d"%(100*p+s),style="font-weight:normal;color:orange",\
                     onclick='Nevow.Athena.Widget.get(this).selected(\'%d\')' % (100*p+s))[ptitle(p,s)]) \
                        or T.div(name=u'selbr%d'%(p))[T.br]) \
                        for p in self.npasses for s in (self.nstages[p]+[999])),
                   ],
              T.td(style="border:thin solid black;")[
                  T.div(name=u"showthumbs")[T.div(name=u"showstage")[
                  T.div(name=u"stagethumbs",style=u'width:800;height:580;overflow:auto')
#                        [
#                        T.table(name=u"tablethumbs")[
#                          T.thead(name=u"tablethumbshead"),
#                          T.tbody(name=u"tablethumbsbody")
#                        ]
#                        ]
                  ]]
              ]
              ]
              ]
        ]
        ]
        return p

    def makeDocFactory( self):
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2(name=u"selheader")[T.div(name=u"seltarget")[self.target]],
               self.table()
             ])

def RecursiveTrue(d):
    if type(d) == TupleType:
       L = d
    else:
       L = d.callbacks
    v = False
    for item in L:
       if type(item) == ListType:
          v=RecursiveUnpause(item)
       elif type(item[0]) == TupleType:
          v=RecursiveUnpause(item)
       else:
          a,b,c = item
          v = hasattr(a,"result")
    return v

def RecursiveUnpause(d):
    if type(d) == TupleType:
       L = d
    else:
       L = d.callbacks
    for item in L:
       if type(item) == ListType:
          RecursiveUnpause(item)
       elif type(item[0]) == TupleType:
          RecursiveUnpause(item)
       else:
          a,b,c = item
          print a, not hasattr(a,"paused") or a.paused
          if hasattr(a,"paused") and a.paused:
             print "Unpausing",a
             a.unpause()

def StartSite(runner,port):
    site = appserver.NevowSite(CIRCEMainPage(runner=runner))
    application = service.Application("CIRCE TV")
    reactor.listenTCP(port,site)

class RemoteBatch:
    def __init__(self,obj,args,batch=100):
        self.obj = obj
        self.args = args
        self.batch = batch
        self.count = 0
    def send(self):
        d = []
        for i in range(self.batch):
           if not len(args): break
           d.append(self.obj.callRemote(*args.pop(0)))
        if len(d) and len(args):
           self.loopcount99 = 0
           self.watch99 = task.LoopingCall(self.notify,self.send,99,1.0,clear=not len(args))
           self.watch99.start(1.0)
    def notify( self,f,n=1,s=0.0,to=60,clear=True):
        dall = [d for d in self.dall if d]
        if not hasattr(self.obj,"looptimes"):
           self.obj.looptimes = {}
        if n in self.obj.looptimes: self.obj.looptimes[n] += 1.0
        else: self.obj.looptimes[n] = 0.0
        exec("self.obj.loopcount%d+= 1"%(n))
        rem = len([d for d in self.dall if not hasattr(d,"result")])
        print "Checking in on",n,len(self.dall),rem,eval("self.obj.loopcount%d"%(n))
        if rem and self.obj.looptimes[n]<to: return
        eval("self.obj.watch%d"%(n)).stop()
        if clear: self.obj.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)


