
from FITS import FITS
from SimpleQueue import *
from basics import *

from twisted.internet import abstract, threads, task, defer
from twisted.python import log as logging
from twisted.application import service, internet
from twisted.python.components import registerAdapter

from nevow               import inevow
from nevow               import accessors
from nevow               import appserver
from nevow               import athena
from nevow               import rend
from nevow               import stan
from nevow               import loaders
from nevow               import url
from nevow               import tags as T
from nevow               import flat
from nevow               import util
from nevow               import livepage
from nevow               import canvas
from nevow               import static

from twisted.internet import task
from twisted.internet import defer, protocol, reactor, address, utils

import time, sys, os
from ephem import formatSexagesimal
from glob import glob

from twisted.python.util import sibpath
from twisted.python.filepath import FilePath

from FourCLiftHelpers import *

TVWATCHERS = []

sibling = FilePath(__file__).sibling

mouseover = u'this.style.cursor="pointer"; this.bgColor="yellow"'
mouseout = u'this.bgColor="white";'

DOCTYPE_XHTML = ( '<!DOCTYPE html>')

def generalError(reason):
    print reason

def callRemote(dummy,obj,func,*args,**kws):
    d = getattr(obj,"callRemote")(func,*args)
    d.addErrback(lambda reason: generalError(reason))
    return d

class RootPage(athena.LivePage):

    def __init__ ( self, *args, **kwargs ):
        self.runner = kwargs["runner"]
        k = len(self.runner.watchers)
        self.runner.watchers[k] = self
        kwargs.pop("runner")
        athena.LivePage.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.showing = u"all"
        self.pdatadir = u"all"
        self.jsModules.mapping[u"EchoThing"] = os.environ["PYTHONDATA"]+"/js/HubSite.js"
        self.disconnected = False

    def locateChild(self,ctx, segments):
        request = inevow.IRequest(ctx)
        if len(segments)>1 and os.path.exists(self.runner.basedir+"/".join(segments)):
           I = None
           suff = segments[-1].split(".")[-1]
           if suff in ["jpg","png","gif"]:
              I = static.File( self.runner.basedir+"/".join(segments),"image/%s" % (suff.replace("jpg","jpeg")))
           elif suff in ["txt","dat"]:
              I = static.File( self.runner.basedir+"/".join(segments),"text/plain")
           elif suff == "fits":
              I = static.File( self.runner.basedir+"/".join(segments),"application/fits")
           if I:
              return I,()
           else:
              return athena.LivePage.locateChild(self,ctx,segments)
        return athena.LivePage.locateChild(self,ctx,segments)

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def beforeRender( self, ctx):
        request = inevow.IRequest(ctx)
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)
        kwargs = request.args
        self.showing = u"all"
        self.pdatadir = kwargs.get("datadir",self.runner.datadirs[0])
#        self.pdatadir = kwargs.get("datadir",u"all")
        if kwargs.get("darks",None):
           self.showing = u"darks"
        elif kwargs.get("flats",None):
           self.showing = u"flats"
        elif kwargs.get("targets",None):
           self.showing = u"targets"
        elif kwargs.get("targs",None):
           self.showing = u"targets"
        elif kwargs.get("all",None):
           self.showing = u"all"

    def disconn( self, reason):
        self.targetsTable.disconn(reason)
        self.targetWatcher.disconn(reason)
        self.pipelineStatus.disconn(reason)
        self.pipelinePerf.disconn(reason)
        self.disconnected = True
        for n in range(100):
           if hasattr(self,"watch%d"%(n)):
              w = getattr(self,"watch%d"%(n))
              if w.running: w.stop()

    def render_ticker(self,ctx,data):
        obj = FrameTicker(runner=self.runner)
        obj.setFragmentParent(self)
        self.frameTicker = obj
        return ctx.tag[obj]

    def render_directory(self,ctx,data):
        obj = PipelineDir(runner=self.runner,par=self)
        obj.setFragmentParent(self)
        self.pipelineDir = obj
        return ctx.tag[obj]

    def render_perform(self,ctx,data):
        obj = PipelinePerf(runner=self.runner,par=self)
        obj.setFragmentParent(self)
        self.pipelinePerf = obj
        return ctx.tag[obj]

    def render_status(self,ctx,data):
        obj = PipelineStatus(runner=self.runner,par=self)
        obj.setFragmentParent(self)
        self.pipelineStatus = obj
        return ctx.tag[obj]

    def render_targets(self,ctx,data):
        gtype = self.showing
        if self.showing == u"targets":
           gcontent = u" (Targets)"
        elif self.showing == u"darks":
           gcontent = u" (Darks)"
        elif self.showing == u"flats":
           gcontent = u" (Flats)"
        elif self.showing == u"all":
           gcontent = u" (Calibrations+Targets)"
        obj = TargetsTable(runner=self.runner,gtype=gtype,gcontent=gcontent,par=self)
        obj.setFragmentParent(self)
        self.targetsTable = obj

        return ctx.tag[obj]

    def render_watcher(self,ctx,data):
        obj = TargetWatcher(runner=self.runner,par=self)
        obj.setFragmentParent(self)
        self.targetWatcher=obj
        return ctx.tag[obj]

    def dropmenu( self ):
        return stan.raw("""<style>/*---- CROSS BROWSER DROPDOWN MENU ----*/
ul#nav {margin: 0 0 0 200px; font-size=10px;}
ul.drop a { display:block; color: #fff;}
ul.drop, ul.drop li, ul.drop ul { list-style: none; margin: 0; padding: 0; border: 1px solid #fff; background: #fff; color: #222;}
ul.drop { position: relative; z-index: 597; float: left; }
ul.drop li { float: left; line-height: 1.3em; vertical-align: middle; zoom: 1; padding: 5px 10px; }
ul.drop li.hover, ul.drop li:hover { position: relative; z-index: 599; cursor: default; background: #1e7c9a; }
ul.drop ul { visibility: hidden; position: absolute; top: 100%; left: 0; z-index: 598}
ul.drop ul li { float: none; }
ul.drop ul ul { top: -2px; left: 100%; }
ul.drop li:hover > ul { visibility: visible }
td.simple {border: thin solid black}
td.targList {border: thin solid black}
th.targList {border: thin solid black}
</style>""")
#targetTableBody {height:300; overflow:auto}
#targetTable {border: thin solid black;}

    def makeDocFactory0( self ):
        return loaders.stan(T.html[
                T.head[
                  T.title["Fourstar Carnegie Lifting Pipeline"],
                  self.dropmenu(),
                  ],
                  T.body[
                     self.directory2()
                  ]
                  ])

    def makeDocFactory( self ):
        return loaders.stan([T.html[
#        return loaders.stan([T.xml(DOCTYPE_XHTML),T.html[
                T.head(render=T.directive('liveglue'))[
                  T.title["Fourstar Carnegie Lifting Pipeline"],
                  T.style[self.dropmenu()],
                  ],
                  T.body[
                     T.table[
                     T.tr[
                       T.td[
                         T.table[
                         T.tbody[
                           T.tr[
                            T.td(_class="simple")[T.h2(style=u'width:480')[
                                "Fourstar Carnegie Lifting Pipeline"]],
                            ],
#                           T.tr[
#                            T.td(style=u"height:160")[T.div(render=T.directive("ticker"))],
#                            ],
                           T.tr[
                            T.td(_class="simple")[T.div(render=T.directive("status"))],
                            ],
                           T.tr[
                            T.td(_class="simple")[T.div(render=T.directive("directory"))],
                            ],
                           T.tr[
                            T.td(_class="simple")[T.div(render=T.directive("perform"))],
                            ],
                           T.tr[
                            T.td(_class="simple")[T.div(render=T.directive("targets"))],
                            ]
                         ]
                         ]
                         ],
                       T.td(style=u"height:720")[T.div(render=T.directive("watcher"))],
                     ],
                     ],
                     ]
                ]])

class FourCLiftMainPage(RootPage):
    def child_( self, ctx):
        return RootPage(runner=self.runner)

class FrameTicker(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        self.par = kwargs["par"]
        kwargs.pop("par")
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.datadirs = self.runner.datadirs
#        self.ifiles = glob(self.runner.datadir+"/image?????.fits")
        self.ifiles = []
        self.chipim = {}
        self.showing = None
        self.disconnected = False

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        self.disconnected = True
        for n in range(100):
           if hasattr(self,"watch%d"%(n)):
              w = getattr(self,"watch%d"%(n))
              if w.running: w.stop()


    def fireItUp( self):
#        self.clock = task.LoopingCall(self.loop)
#        self.clock.start(1)
        reactor.callLater(0.01,self.loop)
        self.i = 0
        self.last = None

    def loop( self,arg=None,st=0):
        if self.disconnected: return
        if st: 
           print "Sleep for",st
           time.sleep(st)
        if hasattr(self.page,"callRemote") and hasattr(self,"datadirs"):
           datadir = self.datadirs[argmax([os.stat(datadir)[8] for datadir in self.datadirs])]
           if self.runner.fclass == "newfirm":
              ifiles = [g for g in glob(datadir+"/image?????.fits") if g not in self.ifiles]
           elif self.runner.fclass == "fourstar":
              ifiles = [g for g in glob(datadir+"/fsr_????_??_c?.fits") if g not in self.ifiles]
           if not ifiles:
              ifiles = self.ifiles
           ifiles.sort()
           if ifiles:
              ifile = ifiles[-1]
           if ifile != self.last:
              if self.runner.fclass == "newfirm":
                  F = FITS(ifile)
                  ra = F["ra"]
                  dec = F["dec"]
                  exptime = F["exptime"]
                  F.close()
                  js = []
                  lfile = ifile
                  for c in [1,2,3,4]:
                      b = os.path.basename(ifile).replace(".fits","c%d.jpg" % (c))
                      js += ["jpegs/"+b]
                  b = os.path.basename(ifile)
              elif self.runner.fclass == "fourstar":
                  F = FITS(ifile)
                  ra = F["ra"]
                  dec = F["dec"]
                  exptime = F["exptime"]
                  F.close()
                  js = []
                  idir = os.path.dirname(ifile)
                  lfile = idir+"/"+"_".join(os.path.basename(ifile).split("_")[:-1])+"_.fits"
                  for c in [1,2,3,4]:
                      b = os.path.basename(lfile).replace(".fits","c%d.jpg" % (c))
                      js += ["jpegs/"+b]
                  b = "_".join(os.path.basename(lfile).split("_")[:-1])

              if self.i:
                 self.callRemote('delText', u'ticker1', u'name')
                 self.callRemote('delText', u'ticker1', u'ra')
                 self.callRemote('delText', u'ticker1', u'dec')
                 self.callRemote('delText', u'ticker1', u'exptime')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%s' % (b), u'name')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%s' % (ra), u'ra')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%s' % (dec), u'dec')
              self.callRemote('addElement', u'td', u'', u'ticker1', u'%.1f' % (exptime), u'exptime')

              data = []
              for c in [1,2,3,4]:
                 ff = lfile.replace(".fits","c%d.fits" % (c))
                 b = os.path.basename(lfile).replace(".fits","c%d.jpg" % (c))
                 j1 = "jpegs/"+b
                 j2 = self.runner.basedir+"/jpegs/"+b
                 if os.path.exists(ff):
                    if c in self.chipim:
                       self.callRemote('delText', u'chip%d'%(c), u'chipim%d'%(c))
                       self.chipim.pop(c)
                    data.append([ff,c,j2])

#               self.wt = task.LoopingCall(self.WatchTicks,data)
#               self.wt = self.wt.start(0.1)

              send = lambda (ff,c,j1): self.TickOneStep2(ff,c,j1)
              Q = RunForkedQueue(send,data,MaxThreads=4,sleeptime=0.001)

              for c in [1,2,3,4]:
#                  if c in self.chipim:
#                     self.callRemote('delText', u'chip%d'%(c), u'chipim%d'%(c))
                  self.callRemote('addImage', u'chip%d'%(c), u'%s'%(j1), u'chipim%d'%(c))
                  self.callRemote('changeElement', u'style', u'width:90', u'chipim%d'%(c))
                  self.chipim[c] = 1

              self.last = ifile
              self.i += 1

        else:
           reactor.callLater(5,self.loop)

    def WatchTicks(self,data):
        if len([d for d in data if os.path.exists(d[2])]) == 4:
           reactor.callLater(2,self.loop)
           reactor.callLater(0,self.wt.stop)

    def TickOneStep2( self,ifile,c,j1):
           print self,ifile,c,j1
           F = FITS(ifile)
           ra = F["ra"]
           dec = F["dec"]
           exptime = F["exptime"]
           if self.runner.fclass == "newfirm":
              ncoadd = F["ncoadd"]
              if ncoadd == "N/A": ncoadd = 1
              F.goto(1+c)
           data = F.data()
           F.close()
           data = data[:2048,:2048]
           data = where(between(data,-1e24,1e24),data,0).astype(float32)
           easydir(self.runner.basedir+"/jpegs/")
           MakeThumbnail(data,mag=16,pct=40,s1=-10,s2=10,inv=1,q=85,j=j1)

           if self.runner.fclass == "newfirm":
               b = os.path.basename(ifile)
           elif self.runner.fclass == "fourstar":
               b = "_".join(os.path.basename(ifile).split("_")[:-1])

#           self.callRemote('addImage', u'chip%d'%(c), u'%s'%(j1), u'chipim%d'%(c))
#           self.callRemote('changeElement', u'style', u'width:90', u'chipim%d'%(c))

    def makeDocFactory( self):
        self.fireItUp()
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2["Frame Ticker"],
             T.table(style="border:thin solid black;")[
             T.thead[ T.tr[
                T.th(style="border:thin solid black;width:90")["Frame"],
                T.th(style="border:thin solid black;width:90")["RA"],
                T.th(style="border:thin solid black;width:90")["DEC"],
                T.th(style="border:thin solid black;width:30")["EXPTIME"],
                ] ],
             T.tbody[
                T.tr(name="ticker1")
                ],
                ],
             T.table(style="border:thin solid black;")[
             T.thead[ T.tr[
                T.th(style="border:thin solid black;width:75")["1"],
                T.th(style="border:thin solid black;width:75")["2"],
                T.th(style="border:thin solid black;width:75")["3"],
                T.th(style="border:thin solid black;width:75")["4"],
                ] ],
             T.tbody[ T.tr(name="ticker2")[
                  T.td[T.div(name=u"chip1")],
                  T.td[T.div(name=u"chip2")],
                  T.td[T.div(name=u"chip3")],
                  T.td[T.div(name=u"chip4")],
                  ] ],
                ],
                ])


class PipelineDir(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        self.par = kwargs["par"]
        kwargs.pop("par")
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.disconnected = False

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def selectd( self, value=None):
        self.par.pdatadir = value
    athena.expose(selectd)

    def makeDocFactory( self):
        click = u'Nevow.Athena.Widget.get(this).selectdir(this.value)'
        sd = [u"all"] + self.runner.datadirs 
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.div[
             T.h2[T.span["Data Directory:"],
             T.select(id="dirSelect", onchange=click)[
               [(datadir == self.par.pdatadir and T.option(selected=1,value=datadir)[datadir] or
                T.option(value=datadir)[datadir]) for datadir in sd]
             ]]
                ],
                ])



class PipelinePerf(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        self.par = kwargs["par"]
        kwargs.pop("par")
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.disconnected = False
        self.pdisp = []
        self.odisp = ""

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        self.disconnected = True
        for n in range(100):
           if hasattr(self,"watch%d"%(n)):
              w = getattr(self,"watch%d"%(n))
              if w.running: w.stop()


    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def check( self):
        reactor.callLater(0.01,self.loop)

    def select( self, value=None):
        if value and value != u"None":
           self.par.targetsTable.select(value)
        else:
           self.callRemote("cursor_clear")
    athena.expose(select)

    def loop( self):
        if not self.disconnected:
           if 1 or self.runner.perform != self.pdisp or self.pdisp==[]:
              if self.runner.perform: pdisp = self.runner.perform
              else: pdisp = ["None"]
              self.pdisp = copy.copy(pdisp)
              tf = lambda t: formatSexagesimal(t/3600.0,4,3600).strip()
              tc = lambda o: self.runner.perfdict[o].current
              ts = lambda o: \
                 (hasattr(self.runner.perfdict[o],"dispatch") and self.runner.perfdict[o].dispatch.Status()[2:]) \
                 or "[]"
              click = lambda o: u'Nevow.Athena.Widget.get(this).selected(\'%s\')' % (o)
              pinfo = lambda o: o!= "None" and \
              "%s [%s] %d %s" % (o,(self.runner.perfdict[o].tfinish>0 and tf(self.runner.perfdict[o].tfinish-self.runner.perfdict[o].tstart) \
                 or tf(time.time()-self.runner.perfdict[o].tstart)),tc(o),ts(o)) or "None"
              html = flat.flatten(T.p[
                 [T.div(onmouseover=mouseover,onmouseout=mouseout)[T.span(onclick=click(o),style=u"color:blue")[pinfo(o)]] for o in pdisp]
              ])
              if html != self.odisp:
                 if self.odisp:
                    dall = [("delText",u"prunning",u"pdisp")]
                 else:
                    dall = []
                 dall += [("addChild",u"prunning",u"pdisp",u"%s"%(html))]
                 self.odisp = html
                 d = RemoteBatch(self,dall,n=96)
                 self.loopcount6 = 0
                 reactor.callLater(2.0,self.notify,[d],self.loop,6)
              else: reactor.callLater(2.0,self.loop)
           else: reactor.callLater(2.0,self.loop)

    def notify( self,dall,f,n=1,s=0.0,to=60):
        dall = [d for d in dall if d]
        if not hasattr(self,"looptimes"):
           self.looptimes = {}
        if n in self.looptimes: self.looptimes[n] += 1.0
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n))
#        if rem == 1:
#           print [d for d in d if not hasattr(d,"result")]
        if rem and self.looptimes[n]<to:
           reactor.callLater(1.0,self.notify,dall,f,n=n,s=s,to=to)
           return
#        w = eval("self.watch%d"%(n))
#        if w.running: w.stop()
        self.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)

    def makeDocFactory( self):
        self.check()
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2[
               T.span["Processing:"],],
             T.pre(name=u"prunning")
                ])



class PipelineStatus(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        self.par = kwargs["par"]
        kwargs.pop("par")
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.disconnected = False

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def check( self):
        reactor.callLater(0.01,self.loop)
#        self.clock = task.LoopingCall(self.loop)
#        self.clock.start(1)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        self.disconnected = True
        for n in range(100):
           if hasattr(self,"watch%d"%(n)):
              w = getattr(self,"watch%d"%(n))
              if w.running: w.stop()

    def loop( self):
        nwatch = [n for n in range(10) if hasattr(self,"watch%d"%(n)) and getattr(self,"watch%d"%(n)).running]
        if self.disconnected: return
        if nwatch:
           print "Remaining watchers",nwatch
           reactor.callLater(1,self.loop)
           return
        d = []
        if hasattr(self.page,"callRemote") and self.runner.going != self.status:
           if self.runner.going:
              status = "Running"
              color = "green"
           else:
              status = "Paused"
              color = "red"
           d.append(self.callRemote('changeElement', u'style', u"color:%s"%(color), u"pstatus"))
           d.append(self.callRemote('delText', u"pstatus", u"status"))
           d.append(self.callRemote('addText', u"span", u"pstatus", u"%s"%(status), u"status"))
        self.status = self.runner.going
        if d:
           self.loopcount3 = 0
           reactor.callLater(1.0,self.notify,d,self.loop,3,1.0)
#           self.watch3 = task.LoopingCall(self.notify,d,self.loop,3,1.0)
#           self.watch3.start(1.0)

    def notify( self,dall,f,n=1,s=0.0,to=60):
        dall = [d for d in dall if d]
        if not hasattr(self,"looptimes"):
           self.looptimes = {}
        if n in self.looptimes: self.looptimes[n] += 1.0
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n))
#        if rem == 1:
#           print [d for d in d if not hasattr(d,"result")]
        if rem and self.looptimes[n]<to:
           reactor.callLater(1.0,self.notify,dall,self.loop,n=n,s=s,to=to)
           return
#        w = eval("self.watch%d"%(n))
#        if w.running: w.stop()
        self.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)

    def select( self, value=None):
        self.runner.toggleGoing()
        self.loop()
    athena.expose(select)

    def makeDocFactory( self):
#        self.check()
        self.status = self.runner.going
        click = u'Nevow.Athena.Widget.get(this).selected(\'switch\')'
        if self.runner.going:
           status = "Running"
           color = "green"
        else:
           status = "Paused"
           color = "red"
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2[
             T.span["Status:"],
             T.span(style=u"color:%s"%(color),name=u"pstatus",onclick=click)[
                T.span(onmouseover=mouseover,onmouseout=mouseout,name=u"status")[status]
                ]
                ],
                ])


class TargetsTable(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        self.gtype = kwargs["gtype"]
        self.gcontent = kwargs["gcontent"]
        self.par = kwargs["par"]
        kwargs.pop("par")
        kwargs.pop("runner")
        kwargs.pop("gtype")
        kwargs.pop("gcontent")
        self.docFactory = self.makeDocFactory()
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.qstats = {}
        self.statuslist = {}
        self.listing = {}
        self.html = None
        self.listtbl = {}
        self.listobj = []
        self.ltime = time.time()
        self.disconnected = False

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def check( self):
#        self.current = None
        self.showing = []
#        reactor.callLater(0.1,self.loop)
#        self.clock = task.LoopingCall(self.loop)
#        self.clock.start(1)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        self.disconnected = True
        for n in range(100):
           if hasattr(self,"watch%d"%(n)):
              w = getattr(self,"watch%d"%(n))
              if w.running: w.stop()

    def getstatus( self, targ):
#         itarg = self.runner.targs.index(targ)+1
         itarg = hash(targ)
         name = u"targ(%d)" % (itarg)
         name1 = u"targ(%d)1" % (itarg)
         name2 = u"targ(%d)2" % (itarg)
         name3 = u"targ(%d)3" % (itarg)
         cname1 = u"ctarg(%d)1" % (itarg)
         cname2 = u"ctarg(%d)2" % (itarg)
         cname3 = u"ctarg(%d)3" % (itarg)
         bname1 = u"btarg(%d)1" % (itarg)
         bname2 = u"btarg(%d)2" % (itarg)
         bname3 = u"btarg(%d)3" % (itarg)
#         self.runner.Data1[targ].Already(first=True,second=False)
#         self.runner.Data2[targ].Already(first=False,second=True)
         done1 = self.runner.Data1[targ].donefirst
         done2 = self.runner.Data2[targ].donesecond
         err1 = self.runner.Data1[targ].errfirst
         err2 = self.runner.Data2[targ].errsecond
         ongoing1 = self.runner.Data1[targ].ongoing
         ongoing2 = self.runner.Data2[targ].ongoing
         bstyle1 = u"padding-left:1px;padding-right:2px;font-weight:normal;"
         bstyle2 = u"padding-left:2px;padding-right:2px;font-weight:normal;text-align:center;"
         bstyle3 = u"padding-left:2px;padding-right:1px;font-weight:normal;"
         status = self.runner.status[targ]
         if err1 or err2:
            bstyle1 += u";color:red"
            bstyle2 += u";color:red"
            bstyle3 += u";color:red"
            status = "Error!"
         elif done1 and not done2:
            bstyle1 += u";color:orange"
            bstyle2 += u";color:orange"
            bstyle3 += u";color:orange"
            if self.runner.second:
               status = self.runner.status[targ]
            else:
               status = "Done..."
         elif done1 and done2:
            bstyle1 += u";color:green"
            bstyle2 += u";color:green"
            bstyle3 += u";color:green"
            status = self.runner.status[targ]
            status = "Done!"
         else:
            bstyle1 += u";color:blue"
            bstyle2 += u";color:blue"
            bstyle3 += u";color:blue"
            status = self.runner.status[targ]
         if ongoing1 or ongoing2:
            bstyle1 += u";font-weight:normal"
            status = "On!"
         else:
            bstyle1 += u";font-weight:normal"
         content1 = u"%s" % (targ)
         content2 = u"%d" % (len(self.runner.groups[targ]))
         content3 = u"%s" % (status)
         click1 = u'Nevow.Athena.Widget.get(this).selected(\'%s\')' % (targ)
         click3 = u'Nevow.Athena.Widget.get(this).selected2(\'%s\')' % (targ)
         return status,done1,done2,ongoing1,ongoing2,name,cname1,cname2,cname3,\
                bstyle1,bstyle2,bstyle3,click1,click3,content1,content2,content3

    def include( self, targ):
        IsItDark = (hasattr(self.runner,"Data1") and self.runner.Data1[targ].type == "dark") or \
                (hasattr(self.runner,"Data2") and self.runner.Data2[targ].type == "dark")
        IsItFlat = (hasattr(self.runner,"Data1") and "flat" in targ and self.runner.Data1[targ].flat) or \
                (hasattr(self.runner,"Data2") and "flat" in targ and self.runner.Data2[targ].flat)
        if self.gtype == u"all":
           return True
        elif self.gtype == u"darks":
           return IsItDark
        elif self.gtype == u"flats":
           return IsItFlat
        elif self.gtype == u"targets":
           return not (IsItDark or IsItFlat)

    def makelist( self):
        showtargs = [targ for targ in self.runner.targs if targ and self.par.pdatadir == u"all" or targ in self.runner.targdirs.get(self.par.pdatadir,[])]
        showtargs.sort()
        showtargs1 = [targ for targ in showtargs if targ and "dark" in targ]
        showtargs1 = [showtargs1[j] for j in argsort([self.runner.groups[targ][0].mjd for targ in showtargs1])]
        showtargs2 = [targ for targ in showtargs if targ and "flat" in targ]
        showtargs2 = [showtargs2[j] for j in argsort([self.runner.groups[targ][0].mjd for targ in showtargs2])]
        showtargs3a = [targ for targ in showtargs if targ and "obj" in targ and "_tmp" not in targ]
        showtargs3a = [showtargs3a[j] for j in argsort([self.runner.groups[targ][0].mjd for targ in showtargs3a])]
        showtargs3 = []
        [(([showtargs3.append(t2) for t2 in showtargs if targ in t2 and "_tmp" in t2] and not showtargs3.append(targ) or showtargs3.append(targ)) \
           and ("sky"+targ[3:] in showtargs and not showtargs3.append("sky"+targ[3:]))) for targ in showtargs3a]
        showtargs4 = [targ for targ in showtargs if targ and "obj" in targ and "tmp" in targ and targ not in showtargs3]
        showtargs4 = [showtargs4[j] for j in argsort([self.runner.groups[targ][0].mjd for targ in showtargs4])]
#        order = sort([self.runner.priorities[targ] for targ in showtargs])
        showtargs = showtargs1 + showtargs2 + showtargs3 + showtargs4
        showtargs = [targ for targ in showtargs if "_tmp" not in targ or targ in self.runner.perform or \
           self.runner.Data1[targ].errfirst or targ in self.runner.left1 or targ.split("_tmp")[0] in self.runner.left1 or \
           (targ.split("_tmp")[0] in self.runner.Data1 and not self.runner.Data1[targ.split("_tmp")[0]].donefirst)]
        rp = {}
        [rp.setdefault( self.runner.priorities[p],p) for p in self.runner.priorities]
        listtbl = {}
        listobj = []
#        for targ in [rp[o] for o in order if rp[o]]:
        for targ in showtargs:
            inc = self.include(targ)
            if inc and targ not in listtbl:
               status,done1,done2,ongoing1,ongoing2,name,cname1,cname2,cname3,\
                   bstyle1,bstyle2,bstyle3,click1,click3,content1,content2,content3 = \
                       self.getstatus(targ)
#               self.statuslist[targ] = status
#               self.qstats[targ] = (done1,done2,ongoing1,ongoing2)
               ne = len(self.runner.groups[targ])
               e1 = self.runner.groups[targ][0].jno
               e2 = self.runner.groups[targ][-1].jno
               basic_info = u'%d [%d,%d]' % (ne,e1,e2)
               p = T.tr(name=name,onmouseover=mouseover,onmouseout=mouseout)[
                 T.td(_class="targList",name=cname1+"d")[
                   T.div(name=cname1,style=bstyle1,onclick=click1)[
                      u'%s' % (targ)
                   ]
                 ],
                 T.td(_class="targList",name=cname2+"d")[
                   T.div(name=cname2,style=bstyle2)[
                      basic_info
                   ]
                 ],
                 T.td(_class="targList",name=cname3+"d")[
                   T.div(name=cname3,style=bstyle3,onclick=click3)[
                      u'%s' % (status)
                   ]
                 ],
               ]
               listtbl[targ] = p
               listobj.append(targ)
               self.listing[targ] = [name,cname1,cname2,cname3]
        return listobj,listtbl

    @defer.inlineCallbacks
    def loop( self):
        nwatch = [n for n in range(10) if hasattr(self,"watch%d"%(n)) and getattr(self,"watch%d"%(n)).running]
        if self.disconnected: return
        dall = []
        ltime = time.time()
#        print "Target table loop! %.3f" % (ltime-self.ltime)
        self.ltime = ltime
        if nwatch:
           print "Remaining watchers",nwatch
           reactor.callLater(2,self.loop)
           return
        if hasattr(self.page,"callRemote"):
           showtargs = [targ for targ in self.runner.targs if self.par.pdatadir == u"all" or targ in self.runner.targdirs.get(self.par.pdatadir,[])]
#           order = sort([self.runner.priorities[targ] for targ in showtargs])
           listobj,listtbl = self.makelist()
           rp = {}
           [rp.setdefault( self.runner.priorities[p],p) for p in self.runner.priorities]
           p = T.table(id="targetTable", name="targetTable", style=u"width:100%")[
               T.thead[ T.tr[
                 T.th(_class="targList")["Grouping"],
                 T.th(_class="targList")["N"],
                 T.th(_class="targList",onclick='Nevow.Athena.Widget.get(this).selected2(\'%s\')' % ('all'))[
                    T.div(onmouseover=mouseover,onmouseout=mouseout)[T.div["Run?"]]
                        ],
                 ] ],
                T.tbody(_id='targetTableBody',name='targetTableBody') [
#                [listtbl[targ] for targ in [rp[o] for o in order if rp[o]] if self.include(targ)]
                [listtbl[targ] for targ in listobj if self.include(targ)]
                ]
                ]
           html = u"%s" % (flat.flatten(p))
           nold = len(self.listobj)
           nnew = len(listobj)
           mold = [j for j in range(min(nold,nnew)) if listobj[j] == self.listobj[j]]
#           print nold,nnew,mold,listobj==self.listobj
           if nold and listobj == self.listobj:
              for targ in listobj:
                  status,done1,done2,ongoing1,ongoing2,name,cname1,cname2,cname3,\
                      bstyle1,bstyle2,bstyle3,click1,click3,content1,content2,content3 = \
                          self.getstatus(targ)
#                  if targ == "obj143824+085400_J1_F55640_3212_tmp0519":
#                     print targ, self.statuslist.get(targ,None), status
                  if targ not in self.statuslist or status != self.statuslist[targ]:
                     self.statuslist[targ] = status
                     self.qstats[targ] = (done1,done2,ongoing1,ongoing2)
                     if targ == self.page.targetWatcher.target:
                       bstyle1 = bstyle1 + u";text-decoration:underline"
                       bstyle2 = bstyle2 + u";text-decoration:underline"
                       bstyle3 = bstyle3 + u";text-decoration:underline"
                     content1 = u"%s" % (targ)
                     content2 = u"%d" % (len(self.runner.groups[targ]))
                     content3 = u"%s" % (status)
                     d1 = ('delText',cname3+"d", cname3)
                     d2 = ('addElement', u'div', bstyle3, cname3+"d", u'%s' % (status), cname3)
                     d3 = ('changeElement', u'style', bstyle1, cname1)
                     d4 = ('changeElement', u'style', bstyle2, cname2)
                     d5 = ('changeElement', u'style', bstyle3, cname3)
                     dall.extend([d1,d2,d3,d4,d5])
                     if u"False" in content3:
                         d6 = ('changeElement', u'onmouseover', mouseover, name)
                         d7 = ('changeElement', u'onmouseout', mouseout, name)
                         d8 = ('changeElement', u'onclick', click1, cname1)
                         d9 = ('changeElement', u'onclick', click3, cname3)
                         dall.extend([d6,d7,d8,d9])
                     yield
              self.listobj = listobj
              self.html = html
           elif self.html != html:
              if self.html:
                 dall += [("delText",u"targetList",u"targetTable")]
              dall += [("addTable",u"targetList",u"targetTable",html)]
              self.listobj = listobj
              self.html = html
        if len(dall):
           d = RemoteBatch(self,dall)
           self.runner.Left(0)
           self.loopcount0 = 0
           reactor.callLater(1.0,self.notify,[d],lambda : self.loop(),0,1.0)
#           self.watch0 = task.LoopingCall(self.notify,[d],lambda : self.loop(),0,1.0)
#           self.watch0.start(1.0)
        else:
           reactor.callLater(2.0,self.loop)

    def notify( self,dall,f,n=1,s=0.0,to=60,clear=True):
        dall = [d for d in dall if d]
        if not hasattr(self,"looptimes"):
           self.looptimes = {}
        if n in self.looptimes: self.looptimes[n] += 1.0
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n))
#        if rem == 1:
#           print [d for d in d if not hasattr(d,"result")]
        if rem and self.looptimes[n]<to:
           reactor.callLater(1.0,self.notify,dall,self.loop,n=n,s=s,to=to,clear=clear)
           return
#        w = eval("self.watch%d"%(n))
#        if w.running: w.stop()
        if clear: self.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)

    def select( self, value=None):
        if value:
           otarg = self.page.targetWatcher.target
           dall = []
           print value,otarg
           if otarg in self.runner.targs and self.par.pdatadir == u"all" or otarg in self.runner.targdirs.get(self.par.pdatadir,[]):
#              itarg = self.runner.targs.index(otarg)+1
              itarg = hash(otarg)
              name = u"targ(%d)" % (itarg)
              name1 = u"ctarg(%d)1" % (itarg)
              dall.append(('changeElement',u"style",u"color:black;text-decoration:normal",name))
#           itarg = self.runner.targs.index(value)+1
           itarg = hash(value)
           name = u"targ(%d)" % (itarg)
           name1 = u"ctarg(%d)1" % (itarg)
           dall.append(('changeElement',u"style",u"color:black;text-decoration:underline",name))
           d = RemoteBatch(self,dall,n=98)
           reactor.callLater(0.1,self.page.targetWatcher.fireUp,value)
           self.loopcount1 = 0
           reactor.callLater(1.0,self.notify,[d],None,1)
#           self.watch1 = task.LoopingCall(self.notify,[d],None,1)
#           self.watch1.start(1.0)
    athena.expose(select)

    def moment( self, dummy="", stuff=""):
        print dummy,stuff

    @defer.inlineCallbacks
    def select2( self, value=None):
        if value and value in self.runner.status:
           targs = [value]
        elif value == u"all":
           if self.gtype == u"all":
              targs = [targ for targ in self.runner.status.keys() if targ]
           elif self.gtype == u"flats":
              targs = [targ for targ in self.runner.status.keys() if targ and "flat" in targ]
           elif self.gtype == u"darks":
              targs = [targ for targ in self.runner.status.keys() if targ and "dark" in targ]
           elif self.gtype == u"targets":
              targs = [targ for targ in self.runner.status.keys() if targ and "dark" not in targ and "flat" not in targ]
        else:
           targs = []
        targs = [targ for targ in targs if self.par.pdatadir == u"all" or targ in self.runner.targdirs.get(self.par.pdatadir,[])]
        for targ in targs:
           if targ in self.listing:
#              itarg = self.runner.targs.index(targ)+1
              itarg = hash(targ)
              done1 = self.runner.Data1[targ].donefirst
              done2 = self.runner.Data2[targ].donesecond
              err1 = self.runner.Data1[targ].errfirst
              err2 = self.runner.Data2[targ].errsecond
              ongoing1 = self.runner.Data1[targ].ongoing
              ongoing2 = self.runner.Data2[targ].ongoing
              if err1 or err2:
                 status = "Error!"
              elif done1 and not done2 and not self.runner.second:
                 status = "Done..."
              elif done1 and done2:
                 status = "Done!"
              else:
                 status = not self.runner.status[targ]
                 self.runner.status[targ] = not self.runner.status[targ]
              print "Status!",targ, status, ongoing1, ongoing2
              if ongoing1 or ongoing2:
                 status = "On!"
              print "Status!",targ, status, ongoing1, ongoing2
#              self.statuslist[targ] = status
              d = self.callRemote("cursor_clear")
           yield
#        reactor.callLater(0.5,self.loop)
    athena.expose(select2)

    def select3( self, value=None):
        if value: self.gtype = value
#        self.loopcount2 = 0
#        reactor.callLater(1.0,self.notify,dall,None,2,clear=False)
#        self.watch2 = task.LoopingCall(self.notify,dall,None,2,clear=False)
#        self.watch2.start(1.0)
    athena.expose(select3)

    def select30( self, value=None):
        if self.gtype == u"all":
           self.gtype = u"targets"
           content = u" (Targets)"
        elif self.gtype == u"targets":
           self.gtype = u"flats"
           content = u" (Flats)"
        elif self.gtype == u"flats":
           self.gtype = u"darks"
           content = u" (Darks)"
        elif self.gtype == u"darks":
           self.gtype = u"all"
           content = u" (Calibrations+Targets)"
        d1 = self.callRemote('delText',u"groups",u"gtype")
        d2 = self.callRemote('addText',u"span",u"groups",content,u"gtype")
        dall = [d1,d2]
        self.loopcount2 = 0
        reactor.callLater(1.0,self.notify,dall,None,2,clear=False)
#        self.watch2 = task.LoopingCall(self.notify,dall,None,2,clear=False)
#        self.watch2.start(1.0)
#    athena.expose(select3)

    def makeDocFactory( self):
        self.check()
        click = u'Nevow.Athena.Widget.get(this).selected3(this.value)'
        Toption = lambda s,value: (s and T.option(value=value,selected='1') or T.option(value=value))
        self.targets = []
        reactor.callLater(0.5,self.loop)
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2[T.div[T.span["Exposure Groups:"],
             T.select(id="dirSelect", onchange=click)[
                Toption(self.gtype==u"all",value='all')["Calibrations+Targets"],
                Toption(self.gtype==u"darks",value='darks')["Darks"],
                Toption(self.gtype==u"flats",value='flats')["Flats"],
                Toption(self.gtype==u"targets",value='targets')["Targets"],
             ]]],
             T.div(style=u'position:relative')[
               T.div(style=u'width:480;height:480;overflow:auto',name=u"targetList")
             ]
                ])


    def makeDocFactory0( self):
        self.check()
        self.targets = []
        reactor.callLater(0.5,self.loop)
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2[T.div(onclick='Nevow.Athena.Widget.get(this).selected3(\'gtype\')')
                [
                  T.span(onmouseover=mouseover,onmouseout=mouseout,name=u"gdirs")[
                      T.span(name=u"gdir")["Exposure Groups:"],
                        ],
                  T.span(onmouseover=mouseover,onmouseout=mouseout,name=u"groups")[
                      T.span(name=u"gtype")[self.gcontent]
                      ]
                      ]
                ],
             T.div(style=u'width:480;height:480;overflow:auto',name="targetList")
#             [
#             T.table(style=u"border:thin solid black",name="targetTable")[
#             T.thead[ T.tr[
#                 T.th(style=u"padding-left:1px;padding-right:2px;border:thin solid black;")["Grouping"],
#                 T.th(style=u"padding-left:2px;padding-right:2px;border:thin solid black;")["N"],
#                 T.th(onclick='Nevow.Athena.Widget.get(this).selected2(\'%s\')' % ('all'),\
#                        style=u"padding-left:2px;padding-right:1px;border:thin solid black;")[
#                    T.div(onmouseover=mouseover,onmouseout=mouseout)[T.div["Run?"]]
#                        ],
#                 ] ],
#             T.tbody(name='targetTableBody')
#                ]
#                ]
                ])

class TargetWatcher(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.runner = kwargs["runner"]
        kwargs.pop("runner")
        self.par = kwargs["par"]
        kwargs.pop("par")
        self.target = "No Target Selected"
        self.showP = 0
        self.showS = 0
        self.schunk = 2

        self.npasses = []
        self.nstages = {0:[1,]}
        self.tstages = {0:["N/A"]}

        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.images = {}
        self.looping = False
        self.disconnected = False

    def cleanUp( self):
        d = []
        d.append(('cursor_wait',))
#        for p in self.npasses:
#            for s in self.nstages[p]:
#                print p,s
#                d.append(('delText',u"sel%d"%(100*p+s), u"sel%dt"%(100*p+s)))
#        for p in self.npasses:
#            for s in self.nstages[p]:
#                d.append(('delText',u"thumbs", u"sel%d"%(100*p+s)))
#            d.append(('delText',u'thumbs',u'selbr%d'%(p)))
        if self.target != "No Target Selected":
           d.append(('delText',u'thumbs',u'sel'))
        return d

    @defer.inlineCallbacks
    def loop( self):
        nwatch = [n for n in range(10) if hasattr(self,"watch%d"%(n)) and getattr(self,"watch%d"%(n)).running]
        if self.disconnected: return
        self.looping = True
#        if self.target != self.newtarget: return
        if nwatch:
           print "Remaining watchers",nwatch
           reactor.callLater(1,self.loop)
           return
        style = u"color:orange"
        d = []
        for p in self.npasses:
            for s in self.nstages[p]:
#                if "orange" in self.sdict[p][s] or (p == self.showP and s == self.showS):
#                   oldstyle = self.sdict[p][s]
#                   tdir = self.gettdir(self.target,p,s)
#                   timages = self.thumbpics(tdir)
#                   if timages: style = u"color:green;font-weight:normal;"
#                   else: style = u"color:orange;font-weight:normal;"
#                   self.sdict[p][s] = style
#                   if p == self.showP and s == self.showS: style = style + u"text-decoration:underline"
#                   else: style = style + u"text-decoration:normal"
#                   if len(timages) > len(self.images) or ("orange" in oldstyle and "green" in style):
#                      d.append(('changeElement',u'style',style,u"sel%dt"%(100*p+s)))
                   dc = self.changeUp1( self.target, p, s)
                   if len(dc):
                      d.extend(dc)
                   yield
        if d:
           d = RemoteBatch(self,d,n=97)
           self.loopcount2 = 0
           reactor.callLater(1.0,self.notify,[d],self.loop,2,1.0)
#           self.watch2 = task.LoopingCall(self.notify,[d],self.loop,2,1.0)
#           self.watch2.start(1.0)
        else:
           reactor.callLater(1,self.loop)

    def addUp1( self, target, p, s):
        ptitle = lambda p,s: u"(%d) %s" % (p,self.tstages[p][s-1])
        o = T.div(name=u"sel%d"%(100*p+s))[
                T.div(name=u"sel%dt"%(100*p+s))
        ]
        return o
#        d = []
#        d1 = ('addElement',u"div",u'',u"thumbs",u'',u"sel%d"%(100*p+s))
#        d2 = ('addText',u"div",u"sel%d"%(100*p+s),ptitle(p,s),u"sel%dt"%(100*p+s))
#        d.extend([d1,d2])
        return d

    def changeUp1( self, target, p, s):
        if p == 1: ef = self.runner.Data1[target].ErrorFile()
        elif p == 2: ef = self.runner.Data2[target].ErrorFile()
        tdir = self.gettdir(target,p,s)
        timages = self.thumbpics(tdir)
#        if ".err" in tdir and timages:
        if ".err" in tdir and os.path.exists(ef):
           style = u"color:red;font-weight:normal;"
        elif timages:
           style = u"color:green;font-weight:normal;"
        else:
           style = u"color:orange;font-weight:normal;"
        if p == self.showP and s == self.showS:
           style = style + u"text-decoration:underline"
        else:
           style = style + u"text-decoration:normal"

        d = []

        if p not in self.ndict:
           self.ndict[p] = {}
        if s not in self.ndict[p]:
           self.ndict[p][s] = ""

        if p not in self.sdict:
           self.sdict[p] = {}
        if s not in self.sdict[p]:
           self.sdict[p][s] = ""

        if self.ndict[p][s] != len(timages):
           ptitle = u"(%d) %s (%d)" % (p,self.tstages[p][s-1],len(timages))
           if self.ndict[p][s] > -1:
              d.append(('delText',u"sel%dt"%(100*p+s),u"num%d"%(100*p+s)))
           d.append(('addText',u"span",u"sel%dt"%(100*p+s),ptitle,u"num%d"%(100*p+s)))
           self.ndict[p][s] = len(timages)

        if style != self.sdict[p][s]:
           d.append( ('changeElement',u"style",style,u"sel%dt"%(100*p+s)))
           d.append( ('changeElement',u'onmouseover',mouseover,u"sel%dt"%(100*p+s)))
           d.append( ('changeElement',u'onmouseout',mouseout,u"sel%dt"%(100*p+s)))
           d.append( ('changeElement',u"onclick",u'Nevow.Athena.Widget.get(this).selected(\'%d\')' % (100*p+s),u"sel%dt"%(100*p+s)))
           self.sdict[p][s] = style

#        print "in change up1",target,p,s,len(d)
        return d

    @defer.inlineCallbacks
    def addUp( self, target,dall):
        d = []
        for p in self.npasses:
            for s in self.nstages[p]:
                d.append(self.addUp1(target, p, s))
                yield
        d.append(T.br)
        html = u"%s" % (flat.flatten(d))
        dall.append(("addChild",u"thumbs",u"sel",html))
#                d.extend(self.addUp1(target, p, s))
#            d.append(('addElement',u'br',u'',u'thumbs',u'',u'selbr%d'%(p)))

    @defer.inlineCallbacks
    def changeUp( self, target,d):
        self.sdict = {}
        self.ndict = {}
        for p in self.npasses:
            self.sdict[p] = {}
            self.ndict[p] = {}
            for s in self.nstages[p]:
                self.ndict[p][s] = -1
                self.sdict[p][s] = None
                d.extend(self.changeUp1(target, p, s))
                yield

    def moment( self, dummy="", stuff=""):
        print dummy,stuff

    def fireUp2( self, target):
        dall = []
        dall.append(('delText', u"selheader", u"seltarget"))
        tstring = u"Target: %s" % (target)
        dall.append(('addText', u"div", u"selheader", tstring, u"seltarget"))
        dall.extend(self.cleanUp())
        self.setUp(target=target)
        self.addUp(target,dall)
        self.changeUp(target,dall)
        d = RemoteBatch(self,dall,n=97)
        self.setTarg(target=target)
        self.loopcount4 = 0
        reactor.callLater(1.0,self.notify,[d],not self.looping and self.loop or None,4)
#        self.watch4 = task.LoopingCall(self.notify,[d],self.loop,4)
#        self.watch4.start(1.0)
        return dall

    def fireUp( self, target):
        print "fireUp!",target
#        [getattr(self,"watch%d"%(n)).stop() for n in range(10) if hasattr(self,"watch%d"%(n)) and getattr(self,"watch%d"%(n)).running]
        dall = []
        self.newtarget = target
        if self.target == target:
           dall.append(self.callRemote('cursor_clear'))
        else:
           if self.images:
              d2 = []
              self.clearsnaps(d2)
              d = RemoteBatch(self,d2)
              self.loopcount5 = 0
              reactor.callLater(1.0,self.notify,[d],lambda: self.fireUp2(target),5)
#              self.watch5 = task.LoopingCall(self.notify,[d],lambda: self.fireUp2(target),5)
#              self.watch5.start(1.0)
           else:
              self.fireUp2(target)
#              dall.extend(self.fireUp2(target))
        return dall

    def notify( self,dall,f,n=1,s=0.0,to=60):
        dall = [d for d in dall if d]
        if not hasattr(self,"looptimes"):
           self.looptimes = {}
        if n in self.looptimes: self.looptimes[n] += 1.0
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n))
#        if rem == 1:
#           print [d for d in d if not hasattr(d,"result")]
        if rem and self.looptimes[n]<to:
           reactor.callLater(1.0,self.notify,dall,f,n=n,s=s,to=to)
           return
#        w = eval("self.watch%d"%(n))
#        if w.running: w.stop()
        self.callRemote("cursor_clear")
        if f:
           reactor.callLater(s,f)


    def setTarg( self, *args, **kws):
        target = kws["target"]
        if target:
           print "Setting target to:",target
           self.target = target

    def setUp( self, *args, **kws):
        target = kws["target"]
        print "In setUp"
        if self.runner.Data1[target].dark:
           self.npasses = [1]
           self.tstages = {1:["Calibrated Frames","Combined Frames","Darks","Error Report"]}
        elif self.runner.Data1[target].flat:
           self.npasses = [1]
           self.tstages = {1:["Calibrated Frames","Count Stats","Combined Frames","Flats","Flat Masks","Error Report"]}
        else:
           self.npasses = [1,2]
           self.tstages = {1:["Calibrated Frames","Count Stats","Simple Sky Frames","Sky Subtracted Frames","Shift Plots",\
                                   "Data Summary","Stacked Frames","Calibration","Mosaicked Frames"],
                           2:["Calibrated Frames","Count Stats","Interpolated Frames","Wavelet Transforms", \
                                   "Wavelet Trends","Model Transforms","Model Sky Frames","Sky Subtracted Frames","Shift Plots",\
                                   "Data Summary","Stacked Frames","Calibration","Mosaicked Frames","Error Report"],
                                   }
           if self.runner.Data1[target].makeflat:
              self.tstages[1] = self.tstages[1] + ["Linearized Frames","Ghost Frames","Sky Flat Frames","Normalized Flat Frames"]
        self.nstages ={}
        for t in self.npasses:
            self.nstages[t] = range(1,len(self.tstages[t])+1)
        self.showP = 0
        self.showS = 0
        self.sdict = {}
        self.ndict = {}

    def select( self, value=None):
        showP = int(value)/100
        showS = int(value)%100

        if self.target in self.runner.targs and value and self.par.pdatadir == u"all" or self.target in self.runner.targdirs.get(self.par.pdatadir,[]):
          name = u"%s" % (self.target.replace(" ",""))
          if hasattr(self,"callRemote"):
             dall = []
#             style = u"text-decoration:underline"
#             item = u"sel%d"%(100*showP+showS)
#             dal.append(('changeElement',u"style",style,item))
             if showP == 1: ef = self.runner.Data1[self.target].ErrorFile()
             elif showP == 2: ef = self.runner.Data2[self.target].ErrorFile()
             tdir = self.gettdir(self.target,showP,showS)
             timages = self.thumbpics(tdir)
#             if ".err" in tdir and timages:
             if ".err" in tdir and os.path.exists(ef):
                style = u"font-weight:normal;color:red;text-decoration:underline"
             elif timages:
                style = u"font-weight:normal;color:green;text-decoration:underline"
             else:
                style = u"font-weight:normal;color:orange;text-decoration:underline"
             item = u"sel%d"%(100*showP+showS)
             dall.append(('changeElement',u"style",style,item))

             dall.append(('delText',u"hproducts",u"products"))
             text = u"Products for %s (Pass %d, %s)" % (name,showP,self.tstages[showP][showS-1])
             dall.append(('addText',u'div',u"hproducts",text,u"products"))
             if self.showP and self.showS:
                tdir = self.gettdir(self.target,self.showP,self.showS)
                timages = self.thumbpics(tdir)
                if timages: style = u"font-weight:normal;color:green;"
                else: style = u"font-weight:normal;color:orange;"
                if showP == self.showP and showS == self.showS:
                   style = style + u"text-decoration:underline"
                else:
                   style = style + u"text-decoration:normal"
                item = u"sel%d"%(100*self.showP+self.showS)
                dall.append(('changeElement',u"style",style,item))
             self.showP = showP
             self.showS = showS
             dall.extend(self.addThumbs())
             d = RemoteBatch(self,dall,n=96)
             self.loopcount9 = 0
             reactor.callLater(1.0,self.notify,[d],None,9)
#             self.watch9 = task.LoopingCall(self.notify,[d],None,9)
#             self.watch9.start(0.5)
        else:
          self.showP = 0
          self.showS = 0
    athena.expose(select)

    def problem(self,reason):
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        self.clock.stop()
        self.disconnected = True
        for n in range(100):
           if hasattr(self,"watch%d"%(n)):
              w = getattr(self,"watch%d"%(n))
              if w.running: w.stop()

    def fireUpProducts( self):
        d = []
        if hasattr(self.page,"callRemote"):
           if self.images:
              self.clearsnaps(d)
           d.extend(self.snapshot())
        return d

    def beforeRender( self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn( self, reason):
        if hasattr(self,"clock") and self.clock.running: self.clock.stop()
        self.disconnected = True
        for n in range(100):
           if hasattr(self,"watch%d"%(n)):
              w = getattr(self,"watch%d"%(n))
              if w.running: w.stop()

    @defer.inlineCallbacks
    def clearsnap( self,pics,j,d):
        for pic in pics[j*self.schunk:(j+1)*self.schunk]:
            lpic = pic.replace("."+pic.split(".")[-1],"link")
            finfo = self.images[pic]
            if 1:
               d.append(( 'delText',lpic,pic))
               d.append(( 'delText',u"thumb%s"%(pic),lpic))
               for k,ftext in enumerate(finfo):
                   if ftext:
                      fwords = ftext.split()
                      nwords = len(fwords)
                      for l in range(nwords):
                          d.append(('delText', u'thumb%sinforow%d' % (pic,k+1), u'thumb%sinfo%d' % (pic,k*nwords+l)))
                          yield
            d.append(( 'removeRows',u"thumb%sinfotbl" % (pic)))
            d.append(( 'delText',u"thumb%sinfo" %(pic),u"thumb%sinfotbl" % (pic)))
            d.append(( 'delText',u"thumb%s" %(pic),lpic))
            d.append(( 'delText',u"thumb%srow1" %(pic),u"thumb%s" % (pic)))
            d.append(( 'delText',u"thumb%srow2" %(pic),u"thumb%sinfo" % (pic)))
            d.append(( 'removeRows',u"thumb%stbl" % (pic)))
            d.append(( 'delText',u"thumb%sitem" %(pic),u"thumb%stbl" % (pic)))
            d.append(( 'delText',u"thumbrow%d" %(j+1),u"thumb%sitem" % (pic)))
            self.images.pop(pic)

    @defer.inlineCallbacks
    def clearsnaps( self,d):
        print "in clearsnaps",len(self.images)
        if self.images:
           print "removing previous images"
           d.append(( 'delText',u"stagethumbs",u"allthumbs"))
           for pic in self.images.keys():
               self.images.pop(pic)
               yield

##           return d
##           d.append(('delText',u"hproducts",u"products"))
##           d.append(('addText',u'div',u"hproducts",u"Clearing Snapshots",u"products"))
#           pics = sorted(self.images.keys())
#           nchunk = (len(pics)-1)/self.schunk + 1
#           for j in range(nchunk):
#               self.clearsnap(pics,j,d)
#               yield
#           d.append( ('removeRows',u"tablethumbs"))
#           d.append( ( 'delText',u"allthumbs",u"tablethumbs"))
#           d.append( ( 'delText',u"stagethumbs",u"allthumbs"))

    def snapshot( self):
#        print "in snapshot"
        d = []
        if hasattr(self.page,"callRemote") and self.showS and self.showP:
#           d.append(('delText',u"hproducts",u"products"))
           name = u"%s" % (self.target.replace(" ",""))
           text = u"Products for %s (Pass %d, %s)" % (name,self.showP,self.tstages[self.showP][self.showS-1])
#           d.append(('addText',u'div',u"hproducts",text,u"products"))
           tdir = self.gettdir(self.target,self.showP,self.showS)
           if u"Stack" in tdir: self.schunk = 3
           elif u"FirstSky" in tdir: self.schunk = 2
           elif u"skyplot" in tdir: self.schunk = 1
           elif u"calphot" in tdir: self.schunk = 1
           elif u"Shift" in tdir: self.schunk = 1
           elif u"shift" in tdir: self.schunk = 3
           elif u"wtrend" in tdir: self.schunk = 2
           elif u"pupil" in tdir: self.schunk = 2
           elif u"Flat" in tdir:
              if u"mode" in tdir:
                 self.schunk = 4
              else:
                 self.schunk = 2
           elif u"Dark" in os.path.basename(tdir):
              if u"mode" in tdir:
                 self.schunk = 4
              else:
                 self.schunk = 2
           elif u"dark" in tdir and u"sky" not in tdir and u"mode" not in tdir: self.schunk = 2
           elif u"mask" in tdir: self.schunk = 2
           else: self.schunk = 4
           p = self.thumbs(tdir)
           html = u"%s" % (flat.flatten(p))
           d = [("addChild",u"stagethumbs",u"allthumbs",html)]
        return d

    def gettdir( self,target,showP,showS):
        if self.runner.fclass == "newfirm": base = "image"
        elif self.runner.fclass == "fourstar": base = "fsr_"
        if showP == 1: ef = self.runner.Data1[target].ErrorFile()
        elif showP == 2: ef = self.runner.Data2[target].ErrorFile()
        else: ef = self.runner.Data1[target].ErrorFile()
        if showP == 1:
           if showS == 1: tdir = u"%s/mode%d/%s" % (target,showP,base)
           elif showS == 2:
               if self.runner.Data1[target].dark:
                  tdir = u"%s/sky%d/Dark" % (target,showP)
#               elif self.runner.Data1[target].flat:
#                  tdir = u"%s/sky%d/Flat" % (target,showP)
               else:
                  tdir = u"%s/skyplot%d" % (target,showP)
           elif showS == 3:
               if self.runner.Data1[target].dark:
                  tdir = u"%s/dark" % (target)
#               elif self.runner.Data1[target].flat:
#                  tdir = u"%s/Flat" % (target)
               elif self.runner.Data1[target].flat:
                  tdir = u"%s/sky%d/Flat" % (target,showP)
               else:
                  tdir = u"%s/sky%d/FirstSky" % (target,showP)
           elif showS == 4:
#               if self.runner.Data1[target].flat:
#                  tdir = u"%s/mask" % (target)
               if self.runner.Data1[target].dark:
                  tdir = u"%s" % (ef)
               if self.runner.Data1[target].flat:
                  tdir = u"%s/Flat" % (target)
               else:
                  tdir = u"%s/sky%d/%s" % (target,showP,base)
           elif showS == 5:
               if self.runner.Data1[target].flat:
                  tdir = u"%s/mask" % (target)
               else:
                  tdir = u"%s/shift%d/%s" % (target,showP,base)
           elif showS == 6:
               if self.runner.Data1[target].flat:
                  tdir = u"%s" % (ef)
               else:
                  tdir = u"%s/Shifts%d" % (target,showP)
           elif showS == 7: tdir = u"%s/Stack%d_" % (target,showP)
           elif showS == 8: tdir = u"%s/calphot%d" % (target,showP)
           elif showS == 9: tdir = u"%s/Mosaic%d" % (target,showP)
           elif showS == 10: tdir = u"%s/mode%d/%s" % (target,0,base)
           elif showS == 11: tdir = u"%s/pupil" % (target,)
           elif showS == 12: tdir = u"%s/sky%d/Flat" % (target,0,)
           elif showS == 13: tdir = u"%s/Flat" % (target,)
        elif showP == 2:
           if showS == 1: tdir = u"%s/mode%d/%s" % (target,showP,base)
           elif showS == 2: tdir = u"%s/skyplot%d" % (target,showP)
           elif showS == 3: tdir = u"%s/back%d/%s" % (target,showP,base)
           elif showS == 4: tdir = u"%s/mall%d/%s" % (target,showP,base)
           elif showS == 5: tdir = u"%s/smow%d/wtrend" % (target,showP)
           elif showS == 6: tdir = u"%s/smow%d/%s" % (target,showP,base)
           elif showS == 7: tdir = u"%s/mbg%d/%s" % (target,showP,base)
           elif showS == 8: tdir = u"%s/sky%d/%s" % (target,showP,base)
           elif showS == 9: tdir = u"%s/shift%d/%s" % (target,showP,base)
           elif showS == 10: tdir = u"%s/Shifts%d" % (target,showP)
           elif showS == 11: tdir = u"%s/Stack%d_" % (target,showP)
           elif showS == 12: tdir = u"%s/calphot%d" % (target,showP)
           elif showS == 13: tdir = u"%s/Mosaic%d" % (target,showP)
           elif showS == 14: tdir = u"%s" % (ef)
#           elif showS == 13: tdir = u"%s.err" % (target)
#        if ".err" in tdir: pass
        if ".err" in tdir and os.path.exists(ef): pass
        elif "dark" in target: tdir = "Darks/"+tdir
        elif "flat" in target: tdir = "Flats/"+tdir
        else:
           bpath = self.runner.Data1[target].zdir
           tdir = "Science/"+bpath+"/"+tdir
        return tdir

    def waitFor(self,d):
        while not d.called:
            print "Waiting on",d,d.called
            time.sleep(1)
        print "Called:",d,d.called

    def dothumb(self,r,j,pic,tdir):
        tdir = os.path.dirname(tdir)+"/"
        kdir = tdir.split("/")[0]
        suff = u"%s" % (pic.split(".")[-1])
        lpic = u"%s" % (pic.replace("."+suff,"link"))
        upic = u"%s" % (tdir+pic.replace("mag16."+suff,"mag01."+suff).replace("mag08."+suff,"mag01."+suff))
        fpic = u"%s" % (pic.replace("mag16","").replace("mag08","").replace("."+suff,""))
        finfo = [
                fpic,
                ""
                ]

        gframe = lambda pdir: u"%s/%s/%s/%s%d/%sc?.fits" % (self.runner.basedir,kdir,self.target,pdir,self.showP,fpic)
        if "shift%d" % (self.showP) in tdir:
           frame = glob(gframe("sky"))
           if frame:
              F = FITS(frame[0])
              xdiff1 = F["xdiff1"]
              ydiff1 = F["ydiff1"]
              mdiff1 = F["mdiff1"]
              xscat1 = F["xscat1"]
              yscat1 = F["yscat1"]
              mscat1 = F["mscat1"]
              lmatch1 = F["lmatch1"]
              F.close()
              finfo = [
                      u"%s %6.2f %6.2f %6.2f" % (fpic,xdiff1,ydiff1,mdiff1),
                      u"%d %6.2f %6.2f %6.2f" % (lmatch1,xscat1,yscat1,mscat1),
                      ]
        elif "mode%d" % (self.showP) in tdir:
           frame = glob(gframe("mode"))
           if frame:
              F = FITS(frame[0])
              medval = F["median"]
              scat = F["scat"]
              F.close()
              finfo = [
                      u"%s" % (fpic),
                      u"median=%8.2f" % (medval),
                      u"sigma=%8.2f" % (scat),
                      ]
        elif "sky%d" % (self.showP) in tdir and "Flat" not in fpic and "Dark" not in fpic:
           frame = glob(gframe("sky"))
           if frame:
              F = FITS(frame[0])
              medval = F["median"]
              anoise = F["anoise"]
              F.close()
              enoise = getsqrt(8*medval + 25**2)/8.
              if "FirstSky" in fpic or "SecondSky" in fpic:
                 finfo = [
                         u"%s" % (fpic),
                         ]
              else:
                 finfo = [
                         u"%s %6.2f %6.2f" % (fpic,anoise,enoise),
                         ]
        else:
           finfo = [
                   fpic,
                   "",
                   ]

        if "Mosaic" in pic:
           width =  u'512'
        elif "Shift" in pic:
           width =  u'512'
        elif "skyplot" in pic:
           width =  u'512'
        elif "calphot" in pic:
           width =  u'512'
        elif "Stack" in pic:
           width =  u'256'
        elif "wtrend" in pic:
           width =  u'512'
        elif "FirstSky" in pic:
           width =  u'256'
        elif "pupil" in pic:
           width =  u'256'
        elif "Flat" in pic:
           width =  u'256'
        elif "Dark" in os.path.basename(pic) or "dark" in os.path.basename(pic):
           width =  u'256'
        elif "mag16" in pic:
           width =  u'128'
        else:
           width =  u'256'

        p = T.td[T.div(name=u'thumb%sitem' % (pic))[
              T.table(name=u'thumb%stbl'%(pic))[
                T.tr(name=u'thumb%srow1'%(pic))[
                  T.td(name=u'thumb%s'%(pic))[
                     T.a(name=lpic,href=upic,target=u"_blank")[T.img(name=pic,src=tdir+pic,width=width)]
                  ]
                ],
                T.tr(name=u'thumb%srow2'%(pic))[
                  T.td(name=u'thumb%sinfo'%(pic))[
                    T.table(name=u'thumb%sinfotbl'%(pic))[
                     [T.tr(name=u'thumb%sinforow%d'%(pic,j+1))[
                      (T.td(name=u"thumb%sinforow%d"%(pic,j*len(ftext.split())+k))[w] for k,w in enumerate(ftext.split()))
                     ] for j,ftext in enumerate(finfo)]
                    ]
                  ]
                ],
              ]
            ]
            ]

        return p,finfo

        d = []
        d.append(('addElement', u'td', u"border:thin solid black;", u'thumbrow%d' % (r+1), u'', u'thumb%sitem' % (pic)))
        d.append(('addElement', u'table', u"", u'thumb%sitem' % (pic), u'', u'thumb%stbl' % (pic)))
        d.append(('addElement', u'tr', u'border: none', u'thumb%stbl' % (pic), u'', u'thumb%srow1' % (pic)))
        d.append(('addElement', u'td', u"", u'thumb%srow1' % (pic), u'', u'thumb%s' % (pic)))
        d.append(('addElement', u'tr', u'border: none', u'thumb%stbl' % (pic), u'', u'thumb%srow2' % (pic)))
        d.append(('addElement', u'td', u"", u'thumb%srow2' % (pic), u'', u'thumb%sinfo' % (pic)))
        d.append(('addElement', u'table', u"", u'thumb%sinfo' % (pic), u'', u'thumb%sinfotbl' % (pic)))

        ninfo = len(finfo)
        for j,ftext in enumerate(finfo):
            if ftext:
               d.append(('addElement', u'tr', u'border: none', u'thumb%sinfotbl' % (pic), u'', u'thumb%sinforow%d' % (pic,j+1)))
               fwords = ftext.split()
               nwords = len(fwords)
               for i in range(nwords):
                   d.append(('addElement', u'td', u"", u'thumb%sinforow%d' % (pic,j+1), u"%s" % (fwords[i]), u'thumb%sinfo%d' % (pic,j*nwords+i)))

        d.append(( 'addElement', u'a', u'', u'thumb%s' % (pic), u'', lpic))

        d.append(( 'changeElement',u'onmouseover',mouseover,u'thumb%sitem' % (pic)))
        d.append(( 'changeElement',u'onmouseout',mouseout,u'thumb%sitem' % (pic)))
        d.append(( 'changeElement', u'href', upic, lpic))
        d.append(( 'changeElement', u'target', u"_blank", lpic))
        d.append(( 'addImage', lpic, tdir+pic, pic))
#        d.append(( 'addImage', u"flow", tdir+pic, pic))
#        d.append(( 'changeElement', u"class", u"item", pic))
#        d.append(( 'changeElement', u"title", pic, pic))
        if "Mosaic" in pic:
           d8 = ( 'changeElement', u'style', u'width:512', pic)
        elif "Shift" in pic:
           d8 = ( 'changeElement', u'style', u'width:512', pic)
        elif "calphot" in pic:
           d8 = ( 'changeElement', u'style', u'width:512', pic)
        elif "skyplot" in pic:
           d8 = ( 'changeElement', u'style', u'width:512', pic)
        elif "Stack" in pic:
           d8 = ( 'changeElement', u'style', u'width:256', pic)
        elif "FirstSky" in pic:
           d8 = ( 'changeElement', u'style', u'width:256', pic)
        elif "Flat" in pic:
           d8 = ( 'changeElement', u'style', u'width:256', pic)
        elif "Dark" in os.path.basename(pic) or "dark" in os.path.basename(pic):
           d8 = ( 'changeElement', u'style', u'width:256', pic)
        elif "mag16" in pic:
           d8 = ( 'changeElement', u'style', u'width:128', pic)
        else:
           d8 = ( 'changeElement', u'style', u'width:256', pic)
        d.append(d8)
        return finfo,d

    def thumbpics(self,tdir):
#        if ".err" in tdir:
#           pics = glob(self.runner.basedir+"/"+tdir)
#        else:
        items1 = glob(self.runner.basedir+"/"+tdir+"*.jpg")
        items2 = glob(self.runner.basedir+"/"+tdir+"*.png")
        bdir = "/".join(tdir.split("/")[:-1]) + "/"
        items = sorted(items1 + items2)
        grab = lambda item: item[-9:] in ["mag16.jpg","mag16.png","mag08.jpg","mag08.png"] or item.split(".")[-1] in ["png"]
        pics = [u"%s" % (os.path.basename(item)) for item in items if grab(item)]
        return pics

    def thumbs(self,tdir):
        if self.showP == 1: ef = self.runner.Data1[self.target].ErrorFile()
        elif self.showP == 2: ef = self.runner.Data2[self.target].ErrorFile()
        if ".err" in tdir and os.path.exists(ef):
           if os.path.exists(ef):
              f = open(ef,"r")
              e = f.read()
              f.close()
              self.images[f.name] = True
              print "Error report exists!"
              p = T.div(name=u'tablethumbs')[
                  T.pre[e]
                  ]
           else:
              p = T.div(name=u'tablethumbs')
              print "Error report does not exist!"
        else:
           pics = self.thumbpics(tdir)
           nchunk = (len(pics)-1)/self.schunk + 1
           bdir = "/".join(tdir.split("/")[:-1]) + "/"
           pr = []
           for j in range(nchunk):
               pd = []
               for p,pic in enumerate(pics[j*self.schunk:(j+1)*self.schunk]):
                   p1,finfo = self.dothumb(j,p+j*self.schunk,pic,bdir)
                   self.images[pic] = finfo
                   pd.append(p1)
               pr.append(pd)
           p = T.table(name=u'tablethumbs')[
              [T.tr[(T.td[pd1] for pd1 in pr1)] for pr1 in pr]
           ]
        return p

    def addThumbs(self):
        d = []
        d.extend(self.fireUpProducts())
        return d

    def table( self):
        name = u"%s" % (self.target.replace(" ",""))
        ptitle = lambda p,s: u"(%d) %s" % (p,self.tstages[p][s-1])
        p = T.div(style=u'position:relative;',name=u"productTable")[
           T.table(style="border:thin solid black;")[
           T.thead[T.tr[
              T.th(style="width:200;border:thin solid black;")["(Pass) Stage"],
              T.th(style="width:800;border:thin solid black;",name=u"hproducts")[T.div(name=u"products")["Select a Stage"]],
              ]],
           T.tbody[T.tr[
              T.td(style="border:thin solid black;",name=u"thumbs")
                [
                   (((s < 999 and T.div(name=u"sel%d"%(100*p+s),style="font-weight:normal;color:orange",\
                     onclick='Nevow.Athena.Widget.get(this).selected(\'%d\')' % (100*p+s))) \
                        or T.div(name=u'selbr%d'%(p))[T.br]) \
                        for p in self.npasses for s in (self.nstages[p]+[999])),
                   ],
              T.td(style="border:thin solid black;")[
                  T.div(name=u"showthumbs")[T.div(name=u"showstage")[
                  T.div(name=u"stagethumbs",style=u'width:800;height:580;overflow:auto')
                  ]]
              ]
              ]
              ]
        ]
        ]
        return p

    def makeDocFactory( self):
        return loaders.stan(T.div(render=T.directive('liveElement'))[
             T.h2(name=u"selheader")[T.div(name=u"seltarget")[self.target]],
               self.table()
             ])

def RecursiveTrue(d):
    if type(d) == TupleType:
       L = d
    else:
       L = d.callbacks
    v = False
    for item in L:
       if type(item) == ListType:
          v=RecursiveUnpause(item)
       elif type(item[0]) == TupleType:
          v=RecursiveUnpause(item)
       else:
          a,b,c = item
          v = hasattr(a,"result")
    return v

def RecursiveUnpause(d):
    if type(d) == TupleType:
       L = d
    else:
       L = d.callbacks
    for item in L:
       if type(item) == ListType:
          RecursiveUnpause(item)
       elif type(item[0]) == TupleType:
          RecursiveUnpause(item)
       else:
          a,b,c = item
          print a, not hasattr(a,"paused") or a.paused
          if hasattr(a,"paused") and a.paused:
             print "Unpausing",a
             a.unpause()

def StartSite(runner,port):
    site = appserver.NevowSite(FourCLiftMainPage(runner=runner))
    application = service.Application("FourCLift TV")
    return reactor.listenTCP(port,site)

class RemoteBatch:
    def __init__(self,obj,args,batch=256,interval=0.1,n=99):
        self.obj = obj
        self.args = args
        self.batch = batch
        self.count = 0
        self.looptimes = {}
        self.int = interval
        self.n = n
        self.start = time.time()
        self.send()
    def done(self):
        self.result = True
    def send(self):
#        d = defer.Deferred()
#        for i in range(self.batch):
#           d.addCallback(lambda result: (len(self.args) and self.obj.callRemote(*self.args.pop(0))) or True)
#        threads.deferToThread(d.callback,True)
#        dall = [d]
        dall = []
        for i in range(self.batch):
           if not len(self.args): break
           dall.append(self.obj.callRemote(*self.args.pop(0)))
        if len(dall):
           setattr(self,"loopcount"+str(self.n),0)
           reactor.callLater(self.int,self.notify,dall,[self.send,self.done][not len(self.args)],self.int)
#           setattr(self,"watch"+str(self.n),task.LoopingCall(self.notify,dall,[self.send,self.done][not len(self.args)],2*self.int))
#           getattr(self,"watch"+str(self.n)).start(self.int)
#           TVWATCHERS.append( getattr(self,"watch"+str(self.n)))
    def notify( self,dall,f,s=0.0,to=60):
        dall = [d for d in dall if d]
        n = self.n
        if n in self.looptimes: self.looptimes[n] = time.time()-self.start
        else: self.looptimes[n] = 0.0
        exec("self.loopcount%d+= 1"%(n))
        rem = len([d for d in dall if not hasattr(d,"result")])
        if eval("self.loopcount%d"%(n)) and not (eval("self.loopcount%d"%(n)) % int(10/self.int)):
           print "Checking in on",n,len(dall),rem,eval("self.loopcount%d"%(n)),self.looptimes[n]
        if rem and self.looptimes[n]<10.0 and hasattr(self.obj,"callRemote"):
           reactor.callLater(self.int,self.notify,dall,f,s=s,to=to)
           return
        print "Do we still have callRemote?",hasattr(self.obj,"callRemote")
#        if getattr(self,"watch"+str(self.n)) in TVWATCHERS:
#           TVWATCHERS.remove( getattr(self,"watch"+str(self.n)))
#           print "Removed",getattr(self,"watch"+str(self.n))
#        eval("self.watch%d"%(n)).stop()
#        if getattr(self,"watch"+str(self.n)).running:
#           getattr(self,"watch"+str(self.n)).stop()
        if f:
           reactor.callLater(s,f)

#style=u"color:%s;font-weight:%s"%(["orange","blue"][datadir==self.runner.pdatadir],["normal","normal"][datadir==self.runner.pdatadir]),selected=(datadir==self.runner.pdatadir)

